from __future__ import absolute_import, division, print_function

import pytest

from odo.numpy_dtype import dshape_to_pandas, unit_to_dtype, dshape_to_numpy
from datashape import dshape
import numpy as np


@pytest.mark.parametrize(
    ['ds', 'expected'],
    [
        ('decimal[9,2]', np.float64),
        ('decimal[9]', np.int32),
        ('?decimal[9]', np.float32),
    ]
)
def test_decimal(ds, expected):
    assert unit_to_dtype(dshape(ds)) == expected


@pytest.mark.parametrize(
    ['ds', 'field'],
    [('var * {teststr1: option[string[4]]}', 'teststr1'),
     ('var * {teststr2: option[string["ascii"]]}', 'teststr2')]
)
def test_parameterized_option_instances(ds, field):
    dtypes, _ = dshape_to_pandas(dshape(ds))
    assert isinstance(dtypes[field], np.dtype)


@pytest.mark.parametrize(
    'ds',
    [
        'option[datetime[tz="EST"]]',
        'option[timedelta[unit="D"]]'
    ]
)
def test_unit_to_dtype(ds):
    assert isinstance(unit_to_dtype(ds), np.dtype)


@pytest.mark.parametrize(
    ['ds', 'expected'],
    [
        ('{a: int32}', ({'a': np.dtype('int32')}, [])),
        ('{a: int32, when: datetime}', ({'a': np.dtype('int32')}, ['when'])),
        ('{a: ?int64}', ({'a': np.dtype('float64')}, []))
    ]
)
def test_dshape_to_pandas(ds, expected):
    assert dshape_to_pandas(ds) == expected


@pytest.mark.parametrize(
    ['ds', 'dt'],
    [
        ('int32', 'int32'),
        ('?int32', 'float32'),
        (
            '{name: string[5, "ascii"], amount: ?int32}',
            [('name', 'S5'), ('amount', '<f4')]
        ),
        ('(int32, float32)', [('f0', '<i4'), ('f1', '<f4')])
    ]
)
def test_dshape_to_numpy(ds, dt):
    assert dshape_to_numpy(ds) == np.dtype(dt)



from __future__ import absolute_import, division, print_function

import os
from bcolz import ctable, carray
import numpy as np
from toolz import keyfilter
import datashape
from datashape import discover
import shutil
from ..numpy_dtype import dshape_to_numpy
from ..append import append
from ..convert import convert, ooc_types
from ..resource import resource
from ..drop import drop
from ..chunks import chunks

keywords = ['cparams', 'dflt', 'expectedlen', 'chunklen', 'rootdir']


@discover.register((ctable, carray))
def discover_bcolz(c, **kwargs):
    return datashape.from_numpy(c.shape, c.dtype)


@append.register((ctable, carray), np.ndarray)
def numpy_append_to_bcolz(a, b, **kwargs):
    a.append(b)
    a.flush()
    return a


@append.register((ctable, carray), object)
def numpy_append_to_bcolz(a, b, **kwargs):
    return append(a, convert(chunks(np.ndarray), b, **kwargs), **kwargs)


@convert.register(ctable, np.ndarray, cost=2.0)
def convert_numpy_to_bcolz_ctable(x, **kwargs):
    return ctable(x, **keyfilter(keywords.__contains__, kwargs))


@convert.register(carray, np.ndarray, cost=2.0)
def convert_numpy_to_bcolz_carray(x, **kwargs):
    return carray(x, **keyfilter(keywords.__contains__, kwargs))


@convert.register(np.ndarray, (carray, ctable), cost=1.0)
def convert_bcolz_to_numpy(x, **kwargs):
    return x[:]


@append.register((carray, ctable), chunks(np.ndarray))
def append_carray_with_chunks(a, c, **kwargs):
    for chunk in c:
        append(a, chunk)
    a.flush()
    return a


@convert.register(chunks(np.ndarray), (ctable, carray), cost=1.2)
def bcolz_to_numpy_chunks(x, chunksize=2**20, **kwargs):
    def load():
        first_n = min(1000, chunksize)
        first = x[:first_n]
        yield first
        for i in range(first_n, x.shape[0], chunksize):
            yield x[i: i + chunksize]
    return chunks(np.ndarray)(load)


@resource.register('.*\.bcolz/?')
def resource_bcolz(uri, dshape=None, expected_dshape=None, **kwargs):
    if os.path.exists(uri):
        try:
            return ctable(rootdir=uri)
        except IOError:  # __rootdirs__ doesn't exist because we aren't a ctable
            return carray(rootdir=uri)
    else:
        if not dshape:
            raise ValueError("Must specify either existing bcolz directory or"
                             " valid datashape")
        dshape = datashape.dshape(dshape)

        dt = dshape_to_numpy(dshape)
        shape_tail = tuple(map(int, dshape.shape[1:]))  # tail of shape
        if dshape.shape[0] == datashape.var:
            shape = (0,) + shape_tail
        else:
            shape = (int(dshape.shape[0]),) + shape_tail

        x = np.empty(shape=shape, dtype=dt)

        kwargs = keyfilter(keywords.__contains__, kwargs)
        expectedlen = kwargs.pop('expectedlen',
                                 int(expected_dshape[0])
                                 if expected_dshape is not None and
                                 isinstance(expected_dshape[0], datashape.Fixed)
                                 else None)

        if datashape.predicates.isrecord(dshape.measure):
            return ctable(x, rootdir=uri, expectedlen=expectedlen, **kwargs)
        else:
            return carray(x, rootdir=uri, expectedlen=expectedlen, **kwargs)


@drop.register((carray, ctable))
def drop_bcolz(b, **kwargs):
    b.flush()
    shutil.rmtree(b.rootdir)


ooc_types |= set((carray, ctable))



from __future__ import absolute_import, division, print_function

import pytest
pytest.importorskip('sqlalchemy')

import os
from decimal import Decimal
from functools import partial

import datashape
from datashape import discover, dshape, float32, float64
from datashape.util.testing import assert_dshape_equal
import numpy as np
import pandas as pd
import sqlalchemy as sa

from odo import convert, append, resource, into, odo, chunks
from odo.backends.sql import (
    dshape_to_table, create_from_datashape, dshape_to_alchemy
)
from odo.utils import tmpfile, raises

from six import string_types


def test_resource():
    sql = resource('sqlite:///:memory:::mytable',
                   dshape='var * {x: int, y: int}')
    assert isinstance(sql, sa.Table)
    assert sql.name == 'mytable'
    assert isinstance(sql.bind, sa.engine.base.Engine)
    assert set(c.name for c in sql.c) == set(['x', 'y'])


def test_append_and_convert_round_trip():
    engine = sa.create_engine('sqlite:///:memory:')
    metadata = sa.MetaData(engine)
    t = sa.Table('bank', metadata,
                 sa.Column('name', sa.String, primary_key=True),
                 sa.Column('balance', sa.Integer))
    t.create()

    data = [('Alice', 1), ('Bob', 2)]
    append(t, data)

    assert convert(list, t) == data


def test_plus_must_have_text():
    with pytest.raises(NotImplementedError):
        resource('redshift+://user:pass@host:1234/db')


def test_resource_on_file():
    with tmpfile('.db') as fn:
        uri = 'sqlite:///' + fn
        sql = resource(uri, 'foo', dshape='var * {x: int, y: int}')
        assert isinstance(sql, sa.Table)

    with tmpfile('.db') as fn:
        uri = 'sqlite:///' + fn
        sql = resource(uri + '::' + 'foo', dshape='var * {x: int, y: int}')
        assert isinstance(sql, sa.Table)


def test_resource_to_engine():
    with tmpfile('.db') as fn:
        uri = 'sqlite:///' + fn
        r = resource(uri)
        assert isinstance(r, sa.engine.Engine)
        assert r.dialect.name == 'sqlite'


def test_resource_to_engine_to_create_tables():
    with tmpfile('.db') as fn:
        uri = 'sqlite:///' + fn
        ds = datashape.dshape('{mytable: var * {name: string, amt: int}}')
        r = resource(uri, dshape=ds)
        assert isinstance(r, sa.engine.Engine)
        assert r.dialect.name == 'sqlite'

        assert discover(r) == ds


def test_discover():
    assert discover(sa.String()) == datashape.string
    metadata = sa.MetaData()
    s = sa.Table('accounts', metadata,
                 sa.Column('name', sa.String),
                 sa.Column('amount', sa.Integer),
                 sa.Column('timestamp', sa.DateTime, primary_key=True))
    ds = dshape('var * {name: ?string, amount: ?int32, timestamp: datetime}')
    assert_dshape_equal(discover(s), ds)
    for name in ds.measure.names:
        assert isinstance(name, string_types)

def test_discover_numeric_column():
    assert discover(sa.String()) == datashape.string
    metadata = sa.MetaData()
    s = sa.Table('name', metadata,
                 sa.Column('name', sa.types.NUMERIC),)

    assert discover(s)


def test_discover_null_columns():
    assert dshape(discover(sa.Column('name', sa.String, nullable=True))) == \
        dshape('{name: ?string}')
    assert dshape(discover(sa.Column('name', sa.String, nullable=False))) == \
        dshape('{name: string}')


def test_discover_selectable():
    t = resource('sqlite:///:memory:::mytable',
                 dshape='var * {x: int, y: int}')
    q = sa.select([t.c.x]).limit(5)
    assert discover(q) == dshape('var * {x: int}')


def test_discover_fixed_length_string():
    t = resource('sqlite:///:memory:::mytable',
                 dshape='var * {x: string[30]}')
    assert discover(t) == dshape('var * {x: string[30]}')


def single_table_engine():
    engine = sa.create_engine('sqlite:///:memory:')
    metadata = sa.MetaData(engine)
    t = sa.Table('accounts', metadata,
                 sa.Column('name', sa.String),
                 sa.Column('amount', sa.Integer))
    t.create()
    return engine, t


def test_select_to_iterator():
    engine, t = single_table_engine()
    append(t, [('Alice', 100), ('Bob', 200)])

    sel = sa.select([t.c.amount + 1])

    assert convert(list, sel) == [(101,), (201,)]
    assert convert(list, sel, dshape=dshape('var * int')) == [101, 201]

    sel2 = sa.select([sa.sql.func.sum(t.c.amount)])

    assert convert(int, sel2, dshape=dshape('int')) == 300

    sel3 = sa.select([t])

    result = convert(list, sel3, dshape=discover(t))
    assert type(result[0]) is tuple
    for res in result:
        assert isinstance(res[0], string_types)


def test_discovery_engine():
    engine, t = single_table_engine()

    assert discover(engine, 'accounts') == discover(t)

    assert str(discover(engine)) == str(discover({'accounts': t}))


def test_discovery_metadata():
    engine, t = single_table_engine()
    metadata = t.metadata
    assert str(discover(metadata)) == str(discover({'accounts': t}))


def test_discover_views():
    engine, t = single_table_engine()
    metadata = t.metadata
    with engine.connect() as conn:
        conn.execute('''CREATE VIEW myview AS
                        SELECT name, amount
                        FROM accounts
                        WHERE amount > 0''')

    assert str(discover(metadata)) == str(
        discover({'accounts': t, 'myview': t}))


def test_extend_empty():
    engine, t = single_table_engine()

    assert not convert(list, t)
    append(t, [])
    assert not convert(list, t)


def test_dshape_to_alchemy():
    assert dshape_to_alchemy('string') == sa.Text
    assert isinstance(dshape_to_alchemy('string[40]'), sa.String)
    assert not isinstance(dshape_to_alchemy('string["ascii"]'), sa.Unicode)
    assert isinstance(dshape_to_alchemy('string[40, "U8"]'), sa.Unicode)
    assert dshape_to_alchemy('string[40]').length == 40

    assert dshape_to_alchemy('float32') == sa.REAL
    assert dshape_to_alchemy('float64') == sa.FLOAT


def test_dshape_to_table():
    t = dshape_to_table('bank', '{name: string, amount: int}')
    assert isinstance(t, sa.Table)
    assert t.name == 'bank'
    assert [c.name for c in t.c] == ['name', 'amount']


td_freqs = list(zip(['D', 'h', 'm', 's', 'ms', 'us', 'ns'],
                    [0, 0, 0, 0, 3, 6, 9],
                    [9, 0, 0, 0, 0, 0, 0]))


@pytest.mark.parametrize(['freq', 'secp', 'dayp'], td_freqs)
def test_dshape_to_table_with_timedelta(freq, secp, dayp):
    ds = '{name: string, amount: int, duration: timedelta[unit="%s"]}' % freq
    t = dshape_to_table('td_bank', ds)
    assert isinstance(t, sa.Table)
    assert t.name == 'td_bank'
    assert isinstance(t.c.duration.type, sa.types.Interval)
    assert t.c.duration.type.second_precision == secp
    assert t.c.duration.type.day_precision == dayp


@pytest.mark.xfail(raises=NotImplementedError)
def test_dshape_to_table_month():
    ds = '{name: string, amount: int, duration: timedelta[unit="M"]}'
    dshape_to_table('td_bank', ds)


@pytest.mark.xfail(raises=NotImplementedError)
def test_dshape_to_table_year():
    ds = '{name: string, amount: int, duration: timedelta[unit="Y"]}'
    dshape_to_table('td_bank', ds)


@pytest.mark.parametrize('freq', ['D', 's', 'ms', 'us', 'ns'])
def test_timedelta_sql_discovery(freq):
    ds = '{name: string, amount: int, duration: timedelta[unit="%s"]}' % freq
    t = dshape_to_table('td_bank', ds)
    assert discover(t).measure['duration'] == datashape.TimeDelta(freq)


@pytest.mark.parametrize('freq', ['h', 'm'])
def test_timedelta_sql_discovery_hour_minute(freq):
    # these always compare equal to a seconds timedelta, because no data loss
    # will occur with this heuristic. this implies that the sa.Table was
    # constructed with day_precision == 0 and second_precision == 0
    ds = '{name: string, amount: int, duration: timedelta[unit="%s"]}' % freq
    t = dshape_to_table('td_bank', ds)
    assert discover(t).measure['duration'] == datashape.TimeDelta('s')


prec = {
    's': 0,
    'ms': 3,
    'us': 6,
    'ns': 9
}


@pytest.mark.parametrize('freq', list(prec.keys()))
def test_discover_postgres_intervals(freq):
    precision = prec.get(freq)
    typ = sa.dialects.postgresql.base.INTERVAL(precision=precision)
    t = sa.Table('t', sa.MetaData(), sa.Column('dur', typ))
    assert discover(t) == dshape('var * {dur: ?timedelta[unit="%s"]}' % freq)


# between postgresql and oracle, only oracle has support for day intervals

@pytest.mark.parametrize('freq', ['D'] + list(prec.keys()))
def test_discover_oracle_intervals(freq):
    typ = sa.dialects.oracle.base.INTERVAL(day_precision={'D': 9}.get(freq),
                                           second_precision=prec.get(freq, 0))
    t = sa.Table('t', sa.MetaData(), sa.Column('dur', typ))
    assert discover(t) == dshape('var * {dur: ?timedelta[unit="%s"]}' % freq)


def test_mssql_types():
    typ = sa.dialects.mssql.BIT()
    t = sa.Table('t', sa.MetaData(), sa.Column('bit', typ))
    assert_dshape_equal(discover(t), dshape('var * {bit: ?bool}'))
    typ = sa.dialects.mssql.DATETIMEOFFSET()
    t = sa.Table('t', sa.MetaData(), sa.Column('dt', typ))
    assert_dshape_equal(discover(t), dshape('var * {dt: ?string}'))
    typ = sa.dialects.mssql.MONEY()
    t = sa.Table('t', sa.MetaData(), sa.Column('money', typ))
    assert_dshape_equal(discover(t), dshape('var * {money: ?float64}'))
    typ = sa.dialects.mssql.SMALLMONEY()
    t = sa.Table('t', sa.MetaData(), sa.Column('money', typ))
    assert_dshape_equal(discover(t), dshape('var * {money: ?float32}'))
    typ = sa.dialects.mssql.UNIQUEIDENTIFIER()
    t = sa.Table('t', sa.MetaData(), sa.Column('uuid', typ))
    assert_dshape_equal(discover(t), dshape('var * {uuid: ?string}'))


def test_create_from_datashape():
    engine = sa.create_engine('sqlite:///:memory:')
    ds = dshape('''{bank: var * {name: string, amount: int},
                    points: var * {x: int, y: int}}''')
    engine = create_from_datashape(engine, ds)

    assert discover(engine) == ds


def test_into_table_iterator():
    engine = sa.create_engine('sqlite:///:memory:')
    metadata = sa.MetaData(engine)
    t = dshape_to_table('points', '{x: int, y: int}', metadata=metadata)
    t.create()

    data = [(1, 1), (2, 4), (3, 9)]
    append(t, data)

    assert convert(list, t) == data
    assert isinstance(convert(list, t)[0], tuple)

    t2 = dshape_to_table('points2', '{x: int, y: int}', metadata=metadata)
    t2.create()
    data2 = [{'x': 1, 'y': 1}, {'x': 2, 'y': 4}, {'x': 3, 'y': 9}]
    append(t2, data2)

    assert convert(list, t2) == data


def test_sql_field_names_disagree_on_order():
    r = resource('sqlite:///:memory:::tb', dshape=dshape('{x: int, y: int}'))
    append(r, [(1, 2), (10, 20)], dshape=dshape('{y: int, x: int}'))
    assert convert(set, r) == set([(2, 1), (20, 10)])


def test_sql_field_names_disagree_on_names():
    r = resource('sqlite:///:memory:::tb', dshape=dshape('{x: int, y: int}'))
    assert raises(Exception, lambda: append(r, [(1, 2), (10, 20)],
                                            dshape=dshape('{x: int, z: int}')))


def test_resource_on_dialects():
    assert (resource.dispatch('mysql://foo') is
            resource.dispatch('mysql+pymysql://foo'))
    assert (resource.dispatch('never-before-seen-sql://foo') is
            resource.dispatch('mysql://foo'))


@pytest.yield_fixture
def sqlite_file():
    try:
        yield 'sqlite:///db.db'
    finally:
        os.remove('db.db')


def test_append_from_select(sqlite_file):
    # we can't test in memory here because that creates two independent
    # databases
    raw = np.array([(200.0, 'Glenn'),
                    (314.14, 'Hope'),
                    (235.43, 'Bob')], dtype=[('amount', 'float64'),
                                             ('name', 'U5')])
    raw2 = np.array([(800.0, 'Joe'),
                     (914.14, 'Alice'),
                     (1235.43, 'Ratso')], dtype=[('amount', 'float64'),
                                                 ('name', 'U5')])
    t = into('%s::t' % sqlite_file, raw)
    s = into('%s::s' % sqlite_file, raw2)
    t = append(t, s.select())
    result = into(list, t)
    expected = np.concatenate((raw, raw2)).tolist()
    assert result == expected


def test_append_from_table():
    # we can't test in memory here because that creates two independent
    # databases
    with tmpfile('db') as fn:
        raw = np.array([(200.0, 'Glenn'),
                        (314.14, 'Hope'),
                        (235.43, 'Bob')], dtype=[('amount', 'float64'),
                                                 ('name', 'U5')])
        raw2 = np.array([(800.0, 'Joe'),
                         (914.14, 'Alice'),
                         (1235.43, 'Ratso')], dtype=[('amount', 'float64'),
                                                     ('name', 'U5')])
        t = into('sqlite:///%s::t' % fn, raw)
        s = into('sqlite:///%s::s' % fn, raw2)
        t = append(t, s)
        result = odo(t, list)
        expected = np.concatenate((raw, raw2)).tolist()
        assert result == expected


def test_engine_metadata_caching():
    with tmpfile('db') as fn:
        engine = resource('sqlite:///' + fn)
        a = resource(
            'sqlite:///' + fn + '::a', dshape=dshape('var * {x: int}'))
        b = resource(
            'sqlite:///' + fn + '::b', dshape=dshape('var * {y: int}'))

        assert a.metadata is b.metadata
        assert engine is a.bind is b.bind


def test_copy_one_table_to_a_foreign_engine():
    data = [(1, 1), (2, 4), (3, 9)]
    ds = dshape('var * {x: int, y: int}')
    with tmpfile('db') as fn1:
        with tmpfile('db') as fn2:
            src = into('sqlite:///%s::points' % fn1, data, dshape=ds)
            tgt = into('sqlite:///%s::points' % fn2,
                       sa.select([src]), dshape=ds)

            assert into(set, src) == into(set, tgt)
            assert into(set, data) == into(set, tgt)


def test_select_to_series_retains_name():
    data = [(1, 1), (2, 4), (3, 9)]
    ds = dshape('var * {x: int, y: int}')
    with tmpfile('db') as fn1:
        points = odo(data, 'sqlite:///%s::points' % fn1, dshape=ds)
        sel = sa.select([(points.c.x + 1).label('x')])
        series = odo(sel, pd.Series)
    assert series.name == 'x'
    assert odo(series, list) == [x + 1 for x, _ in data]


def test_empty_select_to_empty_frame():
    # data = [(1, 1), (2, 4), (3, 9)]
    ds = dshape('var * {x: int, y: int}')
    with tmpfile('db') as fn1:
        points = resource('sqlite:///%s::points' % fn1, dshape=ds)
        sel = sa.select([points])
        df = odo(sel, pd.DataFrame)
    assert df.empty
    assert df.columns.tolist() == ['x', 'y']


def test_discover_foreign_keys():
    with tmpfile('db') as fn:
        products = resource('sqlite:///%s::products' % fn,
                            dshape="""
                                var * {
                                    product_no: int32,
                                    name: ?string,
                                    price: ?float64
                                }
                            """,
                            primary_key=['product_no'])
        expected = dshape("""var * {
                          order_id: int32,
                          product_no: map[int32, {
                            product_no: int32,
                            name: ?string,
                            price: ?float64
                          }],
                          quantity: ?int32
                        }""")
        orders = resource('sqlite:///%s::orders' % fn,
                          dshape=expected,
                          foreign_keys=dict(product_no=products.c.product_no))
        result = discover(orders)
        assert result == expected


def test_invalid_foreign_keys():
    with tmpfile('db') as fn:
        expected = dshape("""var * {
                          order_id: int32,
                          product_no: map[int32, {
                            product_no: int32,
                            name: ?string,
                            price: ?float64
                          }],
                          quantity: ?int32
                        }""")
        with pytest.raises(TypeError):
            resource('sqlite:///%s::orders' % fn, dshape=expected)


def test_foreign_keys_auto_construct():
    with tmpfile('db') as fn:
        products = resource('sqlite:///%s::products' % fn,
                            dshape="""
                                var * {
                                    product_no: int32,
                                    name: ?string,
                                    price: ?float64
                                }
                            """,
                            primary_key=['product_no'])
        ds = dshape("""var * {
                          order_id: int32,
                          product_no: map[int32, T],
                          quantity: ?int32
                        }""")
        orders = resource('sqlite:///%s::orders' % fn, dshape=ds,
                          foreign_keys=dict(product_no=products.c.product_no),
                          primary_key=['order_id'])
        assert discover(orders) == dshape("""
            var * {
                order_id: int32,
                product_no: map[int32, {
                                    product_no: int32,
                                    name: ?string,
                                    price: ?float64
                                }],
                quantity: ?int32
            }
        """)


def test_foreign_keys_bad_field():
    with tmpfile('db') as fn:
        expected = dshape("""var * {
                          order_id: int32,
                          product_no: int64,
                          quantity: ?int32
                        }""")
        with pytest.raises(TypeError):
            resource('sqlite:///%s::orders' % fn, dshape=expected,
                     foreign_keys=dict(foo='products.product_no'))


@pytest.fixture
def recursive_fkey():
    return sa.Table(
        'employees',
        sa.MetaData(),
        sa.Column('eid', sa.BIGINT, primary_key=True),
        sa.Column('name', sa.TEXT),
        sa.Column('mgr_eid', sa.BIGINT, sa.ForeignKey('employees.eid'),
                  nullable=False)
    )


def test_recursive_foreign_key(recursive_fkey):
    expected = dshape("""
        var * {
            eid: int64,
            name: ?string,
            mgr_eid: map[int64, {eid: int64, name: ?string, mgr_eid: int64}]
        }
    """)
    assert discover(recursive_fkey) == expected


def test_create_recursive_foreign_key():
    with tmpfile('.db') as fn:
        t = resource('sqlite:///%s::employees' % fn,
                     dshape="""
                     var * {
                        eid: int64,
                        name: ?string,
                        mgr_eid: map[int64, T]
                     }""", foreign_keys=dict(mgr_eid='employees.eid'),
                     primary_key=['eid'])
        result = discover(t)
    expected = dshape("""
        var * {
            eid: int64,
            name: ?string,
            mgr_eid: map[int64, {eid: int64, name: ?string, mgr_eid: int64}]
        }
    """)
    assert result == expected


def test_compound_primary_key():
    with tmpfile('db') as fn:
        products = resource('sqlite:///%s::products' % fn,
                            dshape="""
                                var * {
                                    product_no: int32,
                                    product_sku: string,
                                    name: ?string,
                                    price: ?float64
                                }
                            """,
                            primary_key=['product_no', 'product_sku'])
        assert len(products.primary_key) == 2
        assert (products.primary_key.columns['product_no'] is
                products.c.product_no)
        assert (products.primary_key.columns['product_sku'] is
                products.c.product_sku)


def test_compound_primary_key_with_fkey():
    with tmpfile('db') as fn:
        products = resource('sqlite:///%s::products' % fn,
                            dshape="""
                                var * {
                                    product_no: int32,
                                    product_sku: string,
                                    name: ?string,
                                    price: ?float64
                                }
                            """,
                            primary_key=['product_no', 'product_sku'])
        ds = dshape("""var * {
                          order_id: int32,
                          product_no: map[int32, T],
                          product_sku: map[int32, U],
                          quantity: ?int32
                        }""")
        orders = resource('sqlite:///%s::orders' % fn, dshape=ds,
                          primary_key=['order_id'],
                          foreign_keys={
                              'product_no': products.c.product_no,
                              'product_sku': products.c.product_sku
                          })
        assert discover(orders) == dshape(
            """var * {
                order_id: int32,
                product_no: map[int32, {product_no: int32, product_sku: string, name: ?string, price: ?float64}],
                product_sku: map[int32, {product_no: int32, product_sku: string, name: ?string, price: ?float64}],
                quantity: ?int32
            }
            """
        )


def test_compound_primary_key_with_single_reference():
    with tmpfile('db') as fn:
        products = resource('sqlite:///%s::products' % fn,
                            dshape="""
                                var * {
                                    product_no: int32,
                                    product_sku: string,
                                    name: ?string,
                                    price: ?float64
                                }
                            """, primary_key=['product_no', 'product_sku'])
        # TODO: should this fail everywhere? e.g., this fails in postgres, but
        # not in sqlite because postgres doesn't allow partial foreign keys
        # might be best to let the backend handle this
        ds = dshape("""var * {
                          order_id: int32,
                          product_no: map[int32, T],
                          quantity: ?int32
                        }""")
        orders = resource('sqlite:///%s::orders' % fn, dshape=ds,
                          foreign_keys=dict(product_no=products.c.product_no),
                          primary_key=['order_id'])
        assert discover(orders) == dshape(
            """var * {
                order_id: int32,
                product_no: map[int32, {product_no: int32, product_sku: string, name: ?string, price: ?float64}],
                quantity: ?int32
            }
            """
        )


def test_foreign_keys_as_compound_primary_key():
    with tmpfile('db') as fn:
        suppliers = resource(
            'sqlite:///%s::suppliers' % fn,
            dshape='var * {id: int64, name: string}',
            primary_key=['id']
        )
        parts = resource(
            'sqlite:///%s::parts' % fn,
            dshape='var * {id: int64, name: string, region: string}',
            primary_key=['id']
        )
        suppart = resource(
            'sqlite:///%s::suppart' % fn,
            dshape='var * {supp_id: map[int64, T], part_id: map[int64, U]}',
            foreign_keys={
                'supp_id': suppliers.c.id,
                'part_id': parts.c.id
            },
            primary_key=['supp_id', 'part_id']
        )
        expected = dshape("""
            var * {
                supp_id: map[int64, {id: int64, name: string}],
                part_id: map[int64, {id: int64, name: string, region: string}]
            }
        """)
        result = discover(suppart)
        assert result == expected


def test_append_chunks():
    tbl = resource('sqlite:///:memory:::test', dshape='var * {a: int, b: int}')
    res = odo(
        chunks(np.ndarray)((
            np.array([[0, 1], [2, 3]]),
            np.array([[4, 5], [6, 7]]),
        )),
        tbl,
    )
    assert res is tbl
    assert (
        odo(tbl, np.ndarray) == np.array(
            [(0, 1),
             (2, 3),
             (4, 5),
             (6, 7)],
            dtype=[('a', '<i4'), ('b', '<i4')],
        )
    ).all()


def test_append_array_without_column_names():
    with pytest.raises(TypeError):
        odo(np.zeros((2, 2)), 'sqlite:///:memory:::test')


def test_numeric_create():
    tbl = resource(
        'sqlite:///:memory:::test',
        dshape='var * {a: ?decimal[11, 2], b: decimal[10, 6]}'
    )
    assert tbl.c.a.nullable
    assert not tbl.c.b.nullable
    assert isinstance(tbl.c.a.type, sa.NUMERIC)
    assert isinstance(tbl.c.b.type, sa.NUMERIC)


def test_numeric_append():
    tbl = resource(
        'sqlite:///:memory:::test',
        dshape='var * {a: decimal[11, 2], b: ?decimal[10, 6]}'
    )
    data = [(1.0, 2.0), (2.0, 3.0)]
    tbl = odo(data, tbl)
    assert odo(tbl, list) == list(map(
        lambda row: tuple(map(Decimal, row)),
        tbl.select().execute().fetchall()
    ))


def test_discover_float_and_real_core_types():
    assert discover(sa.FLOAT()) == float64
    assert discover(sa.REAL()) == float32


def test_string_dshape_doc_example():
    x = np.zeros((10, 2))
    with tmpfile('.db') as fn:
        t = odo(
            x, 'sqlite:///%s::x' % fn, dshape='var * {a: float64, b: float64}'
        )
        assert all(row == (0, 0) for row in t.select().execute().fetchall())


def test_decimal_conversion():
    data = [(1.0,), (2.0,)]
    with tmpfile('.db') as fn:
        t = odo(data, 'sqlite:///%s::x' % fn, dshape='var * {x: decimal[11, 2]}')
        result = odo(sa.select([sa.func.sum(t.c.x)]), Decimal)
    assert result == sum(Decimal(r[0]) for r in data)


def test_append_empty_iterator_returns_table():
    with tmpfile('.db') as fn:
        t = resource('sqlite:///%s::x' % fn, dshape='var * {a: int32}')
        assert odo(iter([]), t) is t


def test_pass_non_hashable_arg_to_create_engine():
    with tmpfile('.db') as fn:
        r = partial(
            resource,
            'sqlite:///%s::t' % fn,
            connect_args={},
            dshape='var * {a: int32}',
        )
        assert r() is r()

    s = partial(
        resource,
        'sqlite:///:memory:::t',
        connect_args={},
        dshape='var * {a: int32}',
    )
    assert s() is not s()



from .into import into


def odo(source, target, **kwargs):
    """ Push one dataset into another

    Parameters
    ----------

    source: object or string
        The source of your data.  Either an object (e.g. DataFrame),
        or a string ('filename.csv')
    target: object or string or type
        The target for where you want your data to go.
        Either an object, (e.g. []), a type, (e.g. list)
        or a string (e.g. 'postgresql://hostname::tablename')
    raise_on_errors: bool (optional, defaults to False)
        Raise exceptions rather than reroute around them
    **kwargs:
        keyword arguments to pass through to conversion functions.

    Optional Keyword Arguments
    --------------------------

    Odo passes keyword arguments (like ``sep=';'``) down to the functions
    that it uses to perform conversions (like ``pandas.read_csv``).  Due to the
    quantity of possible optional keyword arguments we can not list them here.
    See the following documentation for your format

    *   AWS   - http://odo.pydata.org/en/latest/aws.html
    *   CSV   - http://odo.pydata.org/en/latest/csv.html
    *   JSON  - http://odo.pydata.org/en/latest/json.html
    *   HDF5  - http://odo.pydata.org/en/latest/hdf5.html
    *   HDFS  - http://odo.pydata.org/en/latest/hdfs.html
    *   Hive  - http://odo.pydata.org/en/latest/hive.html
    *   SAS   - http://odo.pydata.org/en/latest/sas.html
    *   SQL   - http://odo.pydata.org/en/latest/sql.html
    *   SSH   - http://odo.pydata.org/en/latest/ssh.html
    *   Mongo - http://odo.pydata.org/en/latest/mongo.html
    *   Spark - http://odo.pydata.org/en/latest/spark.html

    Examples
    --------

    >>> L = odo((1, 2, 3), list)  # Convert things into new things
    >>> L
    [1, 2, 3]

    >>> _ = odo((4, 5, 6), L)  # Append things onto existing things
    >>> L
    [1, 2, 3, 4, 5, 6]

    >>> odo([('Alice', 1), ('Bob', 2)], 'myfile.csv')  # doctest: +SKIP

    Explanation
    -----------

    We can specify data with a Python object like a ``list``, ``DataFrame``,
    ``sqlalchemy.Table``, ``h5py.Dataset``, etc..

    We can specify data with a string URI like ``'myfile.csv'``,
    ``'myfiles.*.json'`` or ``'sqlite:///data.db::tablename'``.  These are
    matched by regular expression.  See the ``resource`` function for more
    details on string URIs.

    We can optionally specify datatypes with the ``dshape=`` keyword, providing
    a datashape.  This allows us to be explicit about types when mismatches
    occur or when our data doesn't hold the whole picture.  See the
    ``discover`` function for more information on ``dshape``.

    >>> ds = 'var * {name: string, balance: float64}'
    >>> odo([('Alice', 100), ('Bob', 200)], 'accounts.json', , dshape=ds)  # doctest: +SKIP

    We can optionally specify keyword arguments to pass down to relevant
    conversion functions.  For example, when converting a CSV file we might
    want to specify delimiter

    >>> odo('accounts.csv', list, has_header=True, delimiter=';')  # doctest: +SKIP

    These keyword arguments trickle down to whatever function ``into`` uses
    convert this particular format, functions like ``pandas.read_csv``.

    See Also
    --------

    odo.resource.resource  - Specify things with strings
    datashape.discover      - Get datashape of data
    odo.convert.convert    - Convert things into new things
    odo.append.append      - Add things onto existing things
    """
    return into(target, source, **kwargs)



from __future__ import absolute_import, division, print_function

from collections import Iterator

from toolz import memoize, first, peek
from datashape import discover, var
from .utils import cls_name, copydoc
from dask.threaded import get as dsk_get


class Chunks(object):
    """ An Iterable of chunked data

    Iterates over chunks of in-memory data.  Contains an iterable or a function
    that returns an iterator.

    >>> c = Chunks([[1, 2, 3], [4, 5, 6]])
    >>> next(iter(c))
    [1, 2, 3]

    For typed containers see the ``chunks`` function which generates
    parametrized Chunks classes.

    >>> c = chunks(list)([[1, 2, 3], [4, 5, 6]])
    >>> next(iter(c))
    [1, 2, 3]

    >>> c.container.__name__
    'list'
    """

    def __init__(self, data):
        self.data = data

    def __iter__(self):
        if callable(self.data):
            return self.data()
        elif (isinstance(self.data, list) and
              len(self.data) and
              callable(self.data[0])):
            # If this is a set of callables, evaluate
            # them using dask before returning an iterator for them
            p = []
            dsk = {}
            for i, f in enumerate(self.data):
                dsk['p%d'%i] = (f,)
                p.append('p%d'%i)
            self.data = dsk_get(dsk, p)
        return iter(self.data)


@memoize
@copydoc(Chunks)
def chunks(cls):
    """ Parametrized Chunks Class """
    return type('chunks(%s)' % cls_name(cls), (Chunks,), {'container': cls})


@discover.register(Chunks)
def discover_chunks(c, **kwargs):
    data = c.data
    if isinstance(data, Iterator):
        fst, c.data = peek(data)
    else:
        fst = first(c)
    return var * discover(fst).subshape[0]



from __future__ import absolute_import, division, print_function

import pytest

pymysql = pytest.importorskip('pymysql')

from datashape import var, DataShape, Record, dshape
import itertools
from odo.backends.csv import CSV
from odo import resource, odo
import sqlalchemy
import sqlalchemy as sa
import os
import sys
import csv as csv_module
import getpass
from odo import drop, discover
from odo.utils import tmpfile


pytestmark = pytest.mark.skipif(sys.platform == 'win32',
                                reason='not well tested on win32 mysql')

username = getpass.getuser()
url = 'mysql+pymysql://{0}@localhost:3306/test'.format(username)


def create_csv(data, file_name):
    with open(file_name, 'w') as f:
        csv_writer = csv_module.writer(f)
        for row in data:
            csv_writer.writerow(row)


data = [(1, 2), (10, 20), (100, 200)]
data_floats = [(1.02, 2.02), (102.02, 202.02), (1002.02, 2002.02)]


@pytest.yield_fixture
def csv():
    with tmpfile('.csv') as fn:
        create_csv(data, fn)
        yield CSV(fn)


@pytest.yield_fixture
def fcsv():
    with tmpfile('.csv') as fn:
        create_csv(data_floats, fn)
        yield CSV(fn, columns=list('ab'))


names = ('tbl%d' % i for i in itertools.count())


@pytest.fixture
def name():
    return next(names)


@pytest.fixture(scope='module')
def engine():
    return sqlalchemy.create_engine(url)


@pytest.yield_fixture
def sql(engine, csv, name):
    dshape = var * Record(list(zip('ab', discover(csv).measure.types)))
    try:
        t = resource('%s::%s' % (url, name), dshape=dshape)
    except sqlalchemy.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture
def fsql(engine, fcsv, name):
    dshape = var * Record(list(zip('ab', discover(fcsv).measure.types)))
    try:
        t = resource('%s::%s' % (url, name), dshape=dshape)
    except sqlalchemy.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture
def quoted_sql(engine, fcsv):
    dshape = var * Record(list(zip('ab', discover(fcsv).measure.types)))
    try:
        t = resource('%s::foo bar' % url, dshape=dshape)
    except sqlalchemy.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t
        finally:
            drop(t)


@pytest.fixture
def dcsv():
    return CSV(os.path.join(os.path.dirname(__file__), 'dummydata.csv'))


@pytest.yield_fixture
def dsql(engine, dcsv, name):
    try:
        t = resource('%s::%s' % (url, name), dshape=discover(dcsv))
    except sqlalchemy.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t
        finally:
            drop(t)


@pytest.yield_fixture
def decimal_sql(engine, name):
    try:
        t = resource('%s::%s' % (url, name),
                     dshape="var * {a: ?decimal[10, 3], b: decimal[11, 2]}")
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            yield t
        finally:
            drop(t)



def test_csv_mysql_load(sql, csv):
    engine = sql.bind
    conn = engine.raw_connection()

    cursor = conn.cursor()
    full_path = os.path.abspath(csv.path)
    load = '''LOAD DATA INFILE '{0}' INTO TABLE {1} FIELDS TERMINATED BY ','
        lines terminated by '\n'
        '''.format(full_path, sql.name)
    cursor.execute(load)
    conn.commit()


def test_simple_into(sql, csv):
    odo(csv, sql, if_exists="replace")
    assert odo(sql, list) == [(1, 2), (10, 20), (100, 200)]


def test_append(sql, csv):
    odo(csv, sql, if_exists="replace")
    assert odo(sql, list) == [(1, 2), (10, 20), (100, 200)]

    odo(csv, sql, if_exists="append")
    assert odo(sql, list) == [(1, 2), (10, 20), (100, 200),
                              (1, 2), (10, 20), (100, 200)]


def test_simple_float_into(fsql, fcsv):
    odo(fcsv, fsql, if_exists="replace")

    assert odo(fsql, list) == [(1.02, 2.02),
                               (102.02, 202.02),
                               (1002.02, 2002.02)]


def test_tryexcept_into(sql, csv):
    # uses multi-byte character and fails over to using sql.extend()
    odo(csv, sql, if_exists="replace", QUOTE="alpha", FORMAT="csv")
    assert odo(sql, list) == [(1, 2), (10, 20), (100, 200)]


@pytest.mark.xfail(raises=KeyError)
def test_failing_argument(sql, csv):
    odo(csv, sql, if_exists="replace", skipinitialspace="alpha")


def test_no_header_no_columns(sql, csv):
    odo(csv, sql, if_exists="replace")
    assert odo(sql, list) == [(1, 2), (10, 20), (100, 200)]


@pytest.mark.xfail
def test_complex_into(dsql, dcsv):
    # data from: http://dummydata.me/generate
    odo(dcsv, dsql, if_exists="replace")
    assert odo(dsql, list) == odo(dcsv, list)


def test_sql_to_csv(sql, csv):
    sql = odo(csv, sql)
    with tmpfile('.csv') as fn:
        csv = odo(sql, fn)
        assert odo(csv, list) == data

        # explicitly test that we do NOT preserve the header here
        assert discover(csv).measure.names != discover(sql).measure.names


def test_sql_select_to_csv(sql, csv):
    sql = odo(csv, sql)
    query = sa.select([sql.c.a])
    with tmpfile('.csv') as fn:
        csv = odo(query, fn)
        assert odo(csv, list) == [(x,) for x, _ in data]


def test_csv_output_does_not_preserve_header(sql, csv):
    sql = odo(csv, sql)
    expected = "1,2\n10,20\n100,200\n"
    with tmpfile('.csv') as fn:
        csv = odo(sql, fn)
        with open(csv.path, 'rt') as f:
            result = f.read()
    assert result == expected


@pytest.mark.xfail(raises=AssertionError,
                   reason="Remove when all databases are being tested at once")
def test_different_encoding(name):
    encoding = 'latin1'
    try:
        sql = odo(os.path.join(os.path.dirname(__file__), 'encoding.csv'),
                  url + '::%s' % name,
                  encoding=encoding)
    except sa.exc.OperationalError as e:
        pytest.skip(str(e))
    else:
        try:
            result = odo(sql, list)
            expected = [(u'1958.001.500131-1A', 1, None, u'', 899),
                        (u'1958.001.500156-6', 1, None, u'', 899),
                        (u'1958.001.500162-1', 1, None, u'', 899),
                        (u'1958.001.500204-2', 1, None, u'', 899),
                        (u'1958.001.500204-2A', 1, None, u'', 899),
                        (u'1958.001.500204-2B', 1, None, u'', 899),
                        (u'1958.001.500223-6', 1, None, u'', 9610),
                        (u'1958.001.500233-9', 1, None, u'', 4703),
                        (u'1909.017.000018-3', 1, 30.0, u'sumaria', 899)]
            assert result == expected
        finally:
            drop(sql)


def test_decimal(decimal_sql):
    t = sa.Table(decimal_sql.name, sa.MetaData(decimal_sql.bind), autoload=True)
    assert discover(t) == dshape(
        "var * {a: ?decimal[10, 3], b: decimal[11, 2]}"
    )
    assert isinstance(t.c.a.type, sa.Numeric)
    assert isinstance(t.c.b.type, sa.Numeric)


def test_quoted_name(quoted_sql, fcsv):
    s = odo(fcsv, quoted_sql)
    t = odo(fcsv, list)
    assert sorted(odo(s, list)) == sorted(t)



from __future__ import absolute_import, division, print_function

from odo.append import append

def test_append_list():
    L = [1, 2, 3]
    append(L, [4, 5, 6])
    assert L == [1, 2, 3, 4, 5, 6]


def test_append_list_to_set():
    s = set([1, 2, 3])
    append(s, [4, 5, 6])
    assert s == set([1, 2, 3, 4, 5, 6])


def test_append_set_to_list():
    s = set([3, 4, 5])
    lst = [1, 2, 3]
    append(lst, s)
    assert sorted(lst) == [1, 2, 3, 3, 4, 5]


def test_append_tuple_to_set():
    s = set([1, 2, 3])
    append(s, (3, 4, 5))
    assert s == set([1, 2, 3, 4, 5])



from __future__ import print_function, absolute_import, division

import pytest

pyspark = pytest.importorskip('pyspark')
py4j = pytest.importorskip('py4j')

import os
import shutil
import json
import tempfile
from contextlib import contextmanager

import toolz
from toolz.compatibility import map

from pyspark.sql import Row
from pyspark.sql.types import (
    ArrayType, StructField, StructType, IntegerType, StringType
)

try:
    from pyspark.sql.utils import AnalysisException
except ImportError:
    AnalysisException = py4j.protocol.Py4JJavaError

import numpy as np
import pandas as pd
import pandas.util.testing as tm

import datashape
from datashape import dshape
from odo import odo, discover, Directory, JSONLines
from odo.utils import tmpfile, ignoring
from odo.backends.sparksql import schema_to_dshape, dshape_to_schema
from odo.backends.sparksql import SparkDataFrame


data = [['Alice', 100.0, 1],
        ['Bob', 200.0, 2],
        ['Alice', 50.0, 3]]


df = pd.DataFrame(data, columns=['name', 'amount', 'id'])


@pytest.yield_fixture(scope='module')
def people(sc):
    with tmpfile('.txt') as fn:
        df.to_csv(fn, header=False, index=False)
        raw = sc.textFile(fn)
        parts = raw.map(lambda line: line.split(','))
        yield parts.map(lambda person: Row(name=person[0],
                                           amount=float(person[1]),
                                           id=int(person[2])))


@pytest.fixture(scope='module')
def ctx(sqlctx, people):
    try:
        df = sqlctx.createDataFrame(people)
    except AttributeError:
        schema = sqlctx.inferSchema(people)
        schema.registerTempTable('t')
        schema.registerTempTable('t2')
    else:
        df2 = sqlctx.createDataFrame(people)
        sqlctx.registerDataFrameAsTable(df, 't')
        sqlctx.registerDataFrameAsTable(df2, 't2')
    return sqlctx


def test_pyspark_to_sparksql(ctx, people):
    sdf = odo(data, ctx, dshape=discover(df))
    assert isinstance(sdf, SparkDataFrame)
    assert (list(map(set, odo(people, list))) ==
            list(map(set, odo(sdf, list))))


def test_pyspark_to_sparksql_raises_on_tuple_dshape(ctx, people):
    with pytest.raises(TypeError):
        odo(data, ctx)


def test_dataframe_to_sparksql(ctx):
    sdf = odo(df, ctx)
    assert isinstance(sdf, SparkDataFrame)
    assert odo(sdf, list) == odo(df, list)


def test_sparksql_to_frame(ctx):
    result = odo(ctx.table('t'), pd.DataFrame)
    np.testing.assert_array_equal(result.sort_index(axis=1).values,
                                  df.sort_index(axis=1).values)


def test_reduction_to_scalar(ctx):
    result = odo(ctx.sql('select sum(amount) from t'), float)
    assert isinstance(result, float)
    assert result == sum(map(toolz.second, data))


def test_discover_context(ctx):
    result = discover(ctx)
    assert result is not None


def test_schema_to_dshape():
    assert schema_to_dshape(IntegerType()) == datashape.int32

    assert schema_to_dshape(
        ArrayType(IntegerType(), False)) == dshape("var * int32")

    assert schema_to_dshape(
        ArrayType(IntegerType(), True)) == dshape("var * ?int32")

    assert schema_to_dshape(StructType([
        StructField('name', StringType(), False),
        StructField('amount', IntegerType(), True)])) \
        == dshape("{name: string, amount: ?int32}")


def test_dshape_to_schema():
    assert dshape_to_schema('int32') == IntegerType()

    assert dshape_to_schema('5 * int32') == ArrayType(IntegerType(), False)

    assert dshape_to_schema('5 * ?int32') == ArrayType(IntegerType(), True)

    assert dshape_to_schema('{name: string, amount: int32}') == \
        StructType([StructField('name', StringType(), False),
                    StructField('amount', IntegerType(), False)])

    assert dshape_to_schema('10 * {name: string, amount: ?int32}') == \
        ArrayType(StructType(
            [StructField('name', StringType(), False),
             StructField('amount', IntegerType(), True)]),
        False)


def test_append_spark_df_to_json_lines(ctx):
    out = os.linesep.join(map(json.dumps, df.to_dict('records')))
    sdf = ctx.table('t')
    expected = pd.concat([df, df]).sort('amount').reset_index(drop=True).sort_index(axis=1)
    with tmpfile('.json') as fn:
        with open(fn, mode='w') as f:
            f.write(out + os.linesep)

        uri = 'jsonlines://%s' % fn
        odo(sdf, uri)
        result = odo(uri, pd.DataFrame).sort('amount').reset_index(drop=True).sort_index(axis=1)
        tm.assert_frame_equal(result, expected)


@pytest.mark.xfail(raises=(py4j.protocol.Py4JJavaError, AnalysisException),
                   reason='bug in sparksql')
def test_append(ctx):
    """Add support for odo(SparkDataFrame, SparkDataFrame) when this is fixed.
    """
    a = ctx.table('t2')
    a.insertInto('t2')
    result = odo(odo(a, pd.DataFrame), set)
    expected = odo(pd.concat([odo(a, pd.DataFrame)]) * 2, set)
    assert result == expected


def test_load_from_jsonlines(ctx):
    with tmpfile('.json') as fn:
        js = odo(df, 'jsonlines://%s' % fn)
        result = odo(js, ctx, name='r')
        assert (list(map(set, odo(result, list))) ==
                list(map(set, odo(df, list))))


@contextmanager
def jslines(n=3):
    d = tempfile.mkdtemp()
    files = []
    dfc = df.copy()
    for i in range(n):
        _, fn = tempfile.mkstemp(suffix='.json', dir=d)
        dfc['id'] += i
        odo(dfc, 'jsonlines://%s' % fn)
        files.append(fn)

    yield d

    with ignoring(OSError):
        shutil.rmtree(d)


def test_load_from_dir_of_jsonlines(ctx):
    dfs = []
    dfc = df.copy()
    for i in range(3):
        dfc['id'] += i
        dfs.append(dfc.copy())
    expected = pd.concat(dfs, axis=0, ignore_index=True)
    with jslines() as d:
        result = odo(Directory(JSONLines)(d), ctx)
        assert (set(map(frozenset, odo(result, list))) ==
                set(map(frozenset, odo(expected, list))))



from __future__ import print_function, absolute_import, division

import pytest

pytest.importorskip('pyspark')

import pytest
from datashape import dshape
from odo import odo, discover
from odo.backends.sparksql import SparkDataFrame
from pyspark import RDD
from pyspark.rdd import PipelinedRDD
from pyspark.sql import Row, SchemaRDD


data = [['Alice', 100.0, 1],
        ['Bob', 200.0, 2],
        ['Alice', 50.0, 3]]


@pytest.fixture
def rdd(sc):
    return sc.parallelize(data)


def test_spark_into(rdd):
    with pytest.raises(NotImplementedError):
        odo([1, 2, 3], rdd)


def test_spark_into_context(sc):
    seq = [1, 2, 3]
    rdd = odo(seq, sc)
    assert isinstance(rdd, RDD)
    assert odo(rdd, []) == seq


def test_rdd_into_schema_rdd(rdd):
    ds = dshape('var * {name: string, amount: float64, id: int64}')
    srdd = odo(rdd, SparkDataFrame, dshape=ds)
    assert isinstance(srdd, (SparkDataFrame, SchemaRDD))
    assert list(map(set, srdd.collect())) == list(map(set, rdd.collect()))


def test_pipelined_rdd_into_schema_rdd(rdd):
    pipelined = rdd.map(lambda x: Row(amount=x[1]))
    assert isinstance(pipelined, PipelinedRDD)

    srdd = odo(pipelined, SparkDataFrame,
               dshape=dshape('var * {amount: float64}'))

    assert isinstance(srdd, (SparkDataFrame, SchemaRDD))
    assert (list(map(set, srdd.collect())) ==
            list(map(set, pipelined.collect())))


def test_discover_rdd(rdd):
    assert discover(rdd).subshape[0] == discover(data).subshape[0]



from __future__ import absolute_import, division, print_function

from odo.regex import RegexDispatcher

import re


def test_regex_dispatcher():
    foo = RegexDispatcher('foo')

    @foo.register(r'\d*', priority=9)
    def a(s):
        """ A's docstring """
        return int(s)

    @foo.register(re.compile(r'\D*'))
    def b(s):
        """ B's docstring """
        return s

    @foo.register(r'0\d*', priority=11)
    def c(s):
        """ C's docstring """
        return s

    assert set(foo.funcs.values()) == set([a, b, c])
    assert foo.dispatch('123') == a
    assert foo.dispatch('hello') == b
    assert foo.dispatch('0123') == c

    assert foo('123') == 123
    assert foo('hello') == 'hello'
    assert foo('0123') == '0123'

    assert a.__doc__ in foo.__doc__



from __future__ import absolute_import, division, print_function

import h5py
import os

import datashape
from datashape import DataShape, Record, to_numpy, discover
from datashape.predicates import isrecord
from datashape.dispatch import dispatch
import numpy as np
from toolz import keyfilter

from ..numpy_dtype import dshape_to_numpy
from ..append import append
from ..convert import convert, ooc_types
from ..create import create
from ..resource import resource
from ..chunks import chunks


h5py_attributes = ['chunks', 'compression', 'compression_opts', 'dtype',
                   'fillvalue', 'fletcher32', 'maxshape', 'shape']


try:
    unicode_dtype = h5py.special_dtype(vlen=unicode)
except NameError:
    unicode_dtype = h5py.special_dtype(vlen=str)


@discover.register((h5py.Group, h5py.File))
def discover_h5py_group_file(g):
    return DataShape(Record([[k, discover(v)] for k, v in g.items()]))


def record_dshape_replace(dshape, old, new):
    """Recursively replace all instances of `old` with `new` in the record
    dshape `dshape`.

    Examples
    --------
    >>> from datashape import Record, string, object_, dshape
    >>> ds = DataShape(Record([('a', 'int64'),
    ...                        ('b', 10 * Record([('c', 'object')])),
    ...                        ('d', 'int64')]))
    ...
    >>> Record(list(record_dshape_replace(ds, object_, string)))
    dshape("{a: int64, b: 10 * {c: object}, d: int64}")
    """
    assert isrecord(dshape), 'input dshape must be a record'

    for name, subshape in dshape.measure.fields:
        if subshape == old:
            yield name, new
        else:
            if isrecord(subshape):
                yield record_dshape_replace(subshape, old, new)
            else:
                yield name, subshape


@discover.register(h5py.Dataset)
def discover_h5py_dataset(d):
    dshape = datashape.from_numpy(d.shape, d.dtype)
    shape, measure = dshape.shape, dshape.measure
    if not isrecord(measure):
        if dshape == datashape.object_:
            args = shape + (datashape.string,)
            return DataShape(*args)
        return dshape
    else:
        records = list(record_dshape_replace(measure, datashape.object_,
                                             datashape.string))
        args = shape + (datashape.Record(records),)
        return DataShape(*args)


def dtype_replace(dtype, old, new):
    """Replace the subdtype `old` in `subdtype` with `new`.

    Parameters
    ----------
    dtype, old, new : dtype

    Examples
    --------
    >>> dt = np.dtype([('a', 'int64'), ('b', 'object'),
    ...                ('c', [('d', 'object'), ('e', 'float64')])])
    ...
    >>> r = np.dtype(list(dtype_replace(dt, 'int64', 'float64')))
    >>> r
    dtype([('a', '<f8'), ('b', 'O'), ('c', [('d', 'O'), ('e', '<f8')])])
    """
    names = dtype.names

    assert names is not None, 'dtype must be record-like'

    for name, subdtype in zip(names, map(dtype.__getitem__, names)):
        if subdtype == old:
            yield name, new
        else:
            if subdtype.names is not None:
                yield name, list(dtype_replace(subdtype, old, new))
            else:
                yield name, subdtype


def varlen_dtype(dt):
    """Inject variable length string element for object dtype

    Examples
    --------
    >>> dt = np.dtype('object')
    >>> dt
    dtype('O')
    >>> r = varlen_dtype(dt)
    >>> r
    dtype('O')
    >>> r.metadata['vlen']  # doctest: +SKIP
    <type 'unicode'>
    >>> dt = np.dtype([('a', 'int64'), ('b', 'object'),
    ...                ('c', [('d', 'object'), ('e', 'float64')])])
    ...
    >>> dt['b'].metadata
    >>> r = varlen_dtype(dt)
    >>> r
    dtype([('a', '<i8'), ('b', 'O'), ('c', [('d', 'O'), ('e', '<f8')])])
    >>> r['b'].metadata['vlen']  # doctest: +SKIP
    <type 'unicode'>
    """
    if dt == np.object_:
        return unicode_dtype
    elif dt.names is None:  # some kind of non record like dtype
        return dt
    else:
        return np.dtype(list(dtype_replace(dt, np.dtype('object'),
                                           unicode_dtype)))


def dataset_from_dshape(file, datapath, ds, **kwargs):
    dtype = varlen_dtype(dshape_to_numpy(ds))

    if datashape.var not in list(ds):
        shape = tuple(map(int, ds.shape))
    elif datashape.var not in list(ds)[1:]:
        shape = (0,) + tuple(map(int, ds.shape[1:]))
    else:
        raise ValueError("Don't know how to handle varlen nd shapes")

    if shape:
        kwargs['chunks'] = kwargs.get('chunks', True)
        kwargs['maxshape'] = kwargs.get('maxshape', (None,) + shape[1:])

    kwargs2 = keyfilter(h5py_attributes.__contains__, kwargs)
    return file.require_dataset(datapath, shape=shape, dtype=dtype, **kwargs2)


def create_from_datashape(group, ds, name=None, **kwargs):
    if not isrecord(ds):
        raise ValueError(
            "Trying to create an HDF5 file with non-record datashape failed\n"
            "Perhaps you forgot to specify a datapath?\n"
            "\tdshape: %s\n"
            "If you're using odo consider the following change\n"
            "\tBefore: odo(data, 'myfile.hdf5')\n"
            "\tAfter:  odo(data, 'myfile.hdf5::/datapath')" % ds)
    if isinstance(ds, DataShape) and len(ds) == 1:
        ds = ds[0]
    for name, sub_ds in ds.dict.items():
        if isrecord(sub_ds):
            g = group.require_group(name)
            create_from_datashape(g, sub_ds, **kwargs)
        else:
            dataset_from_dshape(file=group.file,
                                datapath='/'.join([group.name, name]),
                                ds=sub_ds, **kwargs)


@create.register(h5py.File)
def create_h5py_file(cls, path=None, dshape=None, **kwargs):
    f = h5py.File(path)
    create_from_datashape(f, dshape, **kwargs)
    return f


@append.register(h5py.Dataset, np.ndarray)
def append_h5py(dset, x, **kwargs):
    if not sum(x.shape):
        return dset
    shape = list(dset.shape)
    shape[0] += len(x)
    dset.resize(shape)
    dset[-len(x):] = x
    return dset


@append.register(h5py.Dataset, chunks(np.ndarray))
def append_h5py(dset, c, **kwargs):
    for chunk in c:
        append(dset, chunk)
    return dset


@append.register(h5py.Dataset, object)
def append_h5py(dset, x, **kwargs):
    return append(dset, convert(chunks(np.ndarray), x, **kwargs), **kwargs)


@convert.register(np.ndarray, h5py.Dataset, cost=3.0)
def h5py_to_numpy(dset, force=False, **kwargs):
    if dset.size > 1e9:
        raise MemoryError(("File size is large: %0.2f GB.\n"
                           "Convert with flag force=True to force loading") %
                          (dset.size / 1e9))
    else:
        return dset[:]


@convert.register(chunks(np.ndarray), h5py.Dataset, cost=3.0)
def h5py_to_numpy_chunks(dset, chunksize=2 ** 20, **kwargs):
    def load():
        for i in range(0, dset.shape[0], chunksize):
            yield dset[i: i + chunksize]
    return chunks(np.ndarray)(load)


@resource.register('h5py://.+', priority=11)
def resource_h5py(uri, datapath=None, dshape=None, expected_dshape=None,
                  **kwargs):
    if uri.startswith('h5py://'):
        uri = uri[len('h5py://'):]
    f = h5py.File(uri)
    olddatapath = datapath
    if datapath is not None and datapath in f:
        old_dset = f[datapath]
        if expected_dshape is not None:
            dshape = expected_dshape
            assert dshape == discover(old_dset)
    if dshape is not None:
        ds = datashape.dshape(dshape)
        if datapath:
            while ds and datapath:
                datapath, name = datapath.rsplit('/', 1)
                ds = Record([[name, ds]])
            ds = datashape.dshape(ds)
        f.close()
        f = create(h5py.File, path=uri, dshape=ds, **kwargs)
    if olddatapath:
        return f[olddatapath]
    else:
        return f


@resource.register(r'^(?!hdfstore).+\.(hdf5|h5)', priority=10)
def resource_hdf5(uri, *args, **kwargs):
    return resource_h5py(uri, *args, **kwargs)


@dispatch((h5py.Group, h5py.Dataset))
def drop(h):
    del h.file[h.name]


@dispatch(h5py.File)
def drop(h):
    fn = h.filename
    h.close()
    os.remove(fn)


ooc_types.add(h5py.Dataset)



from __future__ import absolute_import, division, print_function

import pytest

pytest.importorskip('dask')

from toolz import merge
from odo.backends.dask import append
from dask.array.core import insert_to_ooc, Array
from dask import core
from odo import convert, into, odo
from odo.utils import tmpfile
import numpy as np


def eq(a, b):
    c = a == b
    if isinstance(c, np.ndarray):
        c = c.all()
    return c


def test_convert():
    x = np.arange(600).reshape((20, 30))
    d = convert(Array, x, chunks=(4, 5))
    assert isinstance(d, Array)


def test_convert_to_numpy_array():
    x = np.arange(600).reshape((20, 30))
    d = convert(Array, x, chunks=(4, 5))
    x2 = convert(np.ndarray, d)
    assert eq(x, x2)


def test_append_to_array():
    bcolz = pytest.importorskip('bcolz')
    x = np.arange(600).reshape((20, 30))
    a = into(Array, x, chunks=(4, 5))
    b = bcolz.zeros(shape=(0, 30), dtype=x.dtype)

    append(b, a)
    assert eq(b[:], x)

    with tmpfile('.hdf5') as fn:
        h = odo(a, fn + '::/data')
        assert eq(h[:], x)
        h.file.close()


def test_into_inplace():
    bcolz = pytest.importorskip('bcolz')
    x = np.arange(600).reshape((20, 30))
    a = into(Array, x, chunks=(4, 5))
    b = bcolz.zeros(shape=(20, 30), dtype=x.dtype)

    append(b, a, inplace=True)
    assert eq(b[:], x)


def test_insert_to_ooc():
    x = np.arange(600).reshape((20, 30))
    y = np.empty(shape=x.shape, dtype=x.dtype)
    a = convert(Array, x, chunks=(4, 5))

    dsk = insert_to_ooc(y, a)
    core.get(merge(dsk, a.dask), list(dsk.keys()))

    assert eq(y, x)


def test_array_interface():
    x = np.arange(600).reshape((20, 30))
    d = convert(Array, x, chunks=(4, 5))

    assert eq(x, np.array(d))



from __future__ import absolute_import, division, print_function

import numpy as np
import datashape as ds
import pytest
tb = pytest.importorskip('tables')

from odo import into, odo
from odo.utils import tmpfile
from odo.backends.pytables import PyTables, discover
import os


try:
    f = tb.open_file('import-tables-test.hdf5', mode='w')
    f.close()
    if os.path.exists('import-tables-test.hdf5'):
        os.remove('import-tables-test.hdf5')
except tb.exceptions.HDF5ExtError as e:
    pytest.skip('Cannot write file, error:\n%s' % e)


x = np.array([(1, 'Alice', 100),
              (2, 'Bob', -200),
              (3, 'Charlie', 300),
              (4, 'Denis', 400),
              (5, 'Edith', -500)],
             dtype=[('id', '<i8'), ('name', 'S7'), ('amount', '<i8')])


@pytest.yield_fixture
def tbfile():
    with tmpfile('.h5') as filename:
        f = tb.open_file(filename, mode='w')
        d = f.create_table('/', 'title',  x)
        d.close()
        f.close()
        yield filename


now = np.datetime64('now').astype('datetime64[us]')
raw_dt_data = [(1, 'Alice', 100, now),
               (2, 'Bob', -200, now),
               (3, 'Charlie', 300, now),
               (4, 'Denis', 400, now),
               (5, 'Edith', -500, now)]


dt_data = np.array(raw_dt_data, dtype=np.dtype([('id', 'i8'),
                                                ('name', 'S7'),
                                                ('amount', 'f8'),
                                                ('date', 'M8[ms]')]))


@pytest.yield_fixture
def dt_tb():
    class Desc(tb.IsDescription):
        id = tb.Int64Col(pos=0)
        name = tb.StringCol(itemsize=7, pos=1)
        amount = tb.Float64Col(pos=2)
        date = tb.Time64Col(pos=3)

    non_date_types = list(zip(['id', 'name', 'amount'], ['i8', 'S7', 'f8']))

    # has to be in microseconds as per pytables spec
    dtype = np.dtype(non_date_types + [('date', 'M8[us]')])
    rec = dt_data.astype(dtype)

    # also has to be a floating point number
    dtype = np.dtype(non_date_types + [('date', 'f8')])
    rec = rec.astype(dtype)
    rec['date'] /= 1e6
    with tmpfile('.h5') as filename:
        f = tb.open_file(filename, mode='w')
        d = f.create_table('/', 'dt', description=Desc)
        d.append(rec)
        d.close()
        f.close()
        yield filename


class TestPyTablesLight(object):

    def test_read(self, tbfile):
        t = PyTables(path=tbfile, datapath='/title')
        shape = t.shape
        t._v_file.close()
        assert shape == (5,)

    def test_write_no_dshape(self, tbfile):
        with pytest.raises(ValueError):
            PyTables(path=tbfile, datapath='/write_this')

    def test_write_with_dshape(self, tbfile):
        f = tb.open_file(tbfile, mode='a')
        try:
            assert '/write_this' not in f
        finally:
            f.close()
            del f

        # create our table
        dshape = '{id: int, name: string[7, "ascii"], amount: float32}'
        t = PyTables(path=tbfile, datapath='/write_this', dshape=dshape)
        shape = t.shape
        filename = t._v_file.filename
        t._v_file.close()

        assert filename == tbfile
        assert shape == (0,)

    @pytest.mark.xfail(reason="Poor datetime support")
    def test_table_into_ndarray(self, dt_tb):
        t = PyTables(dt_tb, '/dt')
        res = into(np.ndarray, t)
        try:
            for k in res.dtype.fields:
                lhs, rhs = res[k], dt_data[k]
                if (issubclass(np.datetime64, lhs.dtype.type) and
                        issubclass(np.datetime64, rhs.dtype.type)):
                    lhs, rhs = lhs.astype('M8[us]'), rhs.astype('M8[us]')
                assert np.array_equal(lhs, rhs)
        finally:
            t._v_file.close()

    def test_ndarray_into_table(self, dt_tb):
        dtype = ds.from_numpy(dt_data.shape, dt_data.dtype)
        t = PyTables(dt_tb, '/out', dtype)
        try:
            res = into(
                np.ndarray, into(t, dt_data, filename=dt_tb, datapath='/out'))
            for k in res.dtype.fields:
                lhs, rhs = res[k], dt_data[k]
                if (issubclass(np.datetime64, lhs.dtype.type) and
                        issubclass(np.datetime64, rhs.dtype.type)):
                    lhs, rhs = lhs.astype('M8[us]'), rhs.astype('M8[us]')
                assert np.array_equal(lhs, rhs)
        finally:
            t._v_file.close()

    @pytest.mark.xfail(reason="Poor datetime support")
    def test_datetime_discovery(self, dt_tb):
        t = PyTables(dt_tb, '/dt')
        lhs, rhs = map(discover, (t, dt_data))
        t._v_file.close()
        assert lhs == rhs

    def test_node_discover(self, dt_tb):
        root = PyTables(dt_tb, '/')
        result = discover(root)
        expected = ds.dshape("""{dt: 5 * {id: int64,
                                          name: string[7, "A"],
                                          amount: float64,
                                          date: float64}}""")
        assert result == expected.measure
        root._v_file.close()

    def test_no_extra_files_around(self, dt_tb):
        """ check the context manager auto-closes the resources """
        assert not len(tb.file._open_files)


def test_pytables_to_csv():
    ndim = 2
    with tmpfile('.h5') as fn:
        h5file = tb.open_file(fn, mode='w', title="Test Array")
        h5file.create_array('/', "test", np.zeros((ndim, ndim), dtype=float))
        h5file.close()
        with tmpfile('csv') as csv:
            t = odo('pytables://%s::/test' % fn, csv)
            assert odo(t, list) == [(0.0, 0.0), (0.0, 0.0)]



from __future__ import absolute_import, division, print_function

from multipledispatch import Dispatcher
from datashape.dispatch import namespace
from .convert import convert


if 'append' not in namespace:
    namespace['append'] = Dispatcher('append')
append = namespace['append']


@append.register(object, object)
def append_not_found(a, b, **kwargs):
    """ Append one dataset on to another

    Examples
    --------

    >>> data = [1, 2, 3]
    >>> _ = append(data, (4, 5, 6))
    >>> data
    [1, 2, 3, 4, 5, 6]
    """
    raise NotImplementedError("Don't know how to append datasets of type "
            "%s on to type %s" % (type(b), type(a)))


@append.register(list, list)
def list_to_list(a, b, **kwargs):
    a.extend(b)
    return a


@append.register(list, object)
def object_to_list(a, b, **kwargs):
    append(a, convert(list, b))
    return a


@append.register(set, set)
def set_to_set(a, b, **kwargs):
    a.update(b)
    return a


@append.register(set, object)
def object_to_set(a, b, **kwargs):
    append(a, convert(set, b))
    return a



import os
import sys
import shutil
import pytest


@pytest.fixture(scope='session')
def sc():
    pytest.importorskip('pyspark')
    from pyspark import SparkContext
    return SparkContext('local[*]', 'odo')


@pytest.yield_fixture(scope='session')
def sqlctx(sc):
    pytest.importorskip('pyspark')
    from odo.backends.sparksql import HiveContext

    try:
        yield HiveContext(sc)
    finally:
        dbpath = 'metastore_db'
        logpath = 'derby.log'
        if os.path.exists(dbpath):
            assert os.path.isdir(dbpath)
            shutil.rmtree(dbpath)
        if os.path.exists(logpath):
            assert os.path.isfile(logpath)
            os.remove(logpath)



from __future__ import absolute_import, division, print_function

from multipledispatch import Dispatcher
import datashape
import numpy as np

create = Dispatcher('create')


@create.register(type)
def create_type(cls, **kwargs):
    if 'dshape' in kwargs:
        kwargs['dshape'] = datashape.dshape(kwargs['dshape'])
    func = create.dispatch(cls)
    return func(cls, **kwargs)


@create.register(np.ndarray)
def create_np_ndarray(_, dshape=None, **kwargs):
    shape, dtype = datashape.to_numpy(dshape)
    return np.empty(shape=shape, dtype=dtype)



from __future__ import absolute_import, division, print_function

import sys
import itertools
import os

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3
ON_TRAVIS_CI = 'TRAVIS_PYTHON_VERSION' in os.environ

if sys.version_info[0] >= 3:
    unicode = str
    map = map
    range = range

    from io import StringIO
else:
    unicode = unicode
    map = itertools.imap
    range = xrange

    from io import BytesIO as StringIO


def skipif(cond, **kwargs):
    def _(func):
        if cond:
            return None
        else:
            return func
    return _

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen



from __future__ import absolute_import, division, print_function

from datetime import datetime, timedelta
from functools import partial

from datashape import discover, Categorical, DateTime
from datashape import string, object_, datetime_, Option
import datashape

import pandas as pd
import numpy as np

from ..convert import convert


possibly_missing = frozenset({string, datetime_})
categorical = type(pd.Categorical.dtype)


def dshape_from_pandas(col):
    if isinstance(col.dtype, categorical):
        return Categorical(col.cat.categories.tolist())
    elif col.dtype.kind == 'M':
        tz = getattr(col.dtype, 'tz', None)
        if tz is not None:
            # Pandas stores this as a pytz.tzinfo, but DataShape wants a
            # string.
            tz = str(tz)
        return Option(DateTime(tz=tz))

    dshape = datashape.CType.from_numpy_dtype(col.dtype)
    dshape = string if dshape == object_ else dshape
    return Option(dshape) if dshape in possibly_missing else dshape


@discover.register(pd.DataFrame)
def discover_dataframe(df):
    return len(df) * datashape.Record([(k, dshape_from_pandas(df[k]))
                                       for k in df.columns])


@discover.register(pd.Series)
def discover_series(s):
    return len(s) * dshape_from_pandas(s)


def coerce_datetimes(df):
    """ Make object columns into datetimes if possible

    Warning: this operates inplace.

    Example
    -------

    >>> df = pd.DataFrame({'dt': ['2014-01-01'], 'name': ['Alice']})
    >>> df.dtypes  # note that these are strings/object
    dt      object
    name    object
    dtype: object

    >>> df2 = coerce_datetimes(df)
    >>> df2
              dt   name
    0 2014-01-01  Alice

    >>> df2.dtypes  # note that only the datetime-looking-one was transformed
    dt      datetime64[ns]
    name            object
    dtype: object
    """
    objects = df.select_dtypes(include=['object'])
    # NOTE: In pandas < 0.17, pd.to_datetime(' ') == datetime(...), which is
    # not what we want.  So we have to remove columns with empty or
    # whitespace-only strings to prevent erroneous datetime coercion.
    columns = [
        c for c in objects.columns
        if not np.any(objects[c].str.isspace() | objects[c].str.isalpha())
    ]
    df2 = objects[columns].apply(partial(pd.to_datetime, errors='ignore'))

    for c in df2.columns:
        df[c] = df2[c]
    return df


@convert.register(pd.Timestamp, datetime)
def convert_datetime_to_timestamp(dt, **kwargs):
    return pd.Timestamp(dt)


@convert.register((pd.Timestamp, pd.Timedelta), float)
def nan_to_nat(fl, **kwargs):
    try:
        if np.isnan(fl):
            # Only nan->nat edge
            return pd.NaT
    except TypeError:
        pass
    raise NotImplementedError()


@convert.register((pd.Timestamp, pd.Timedelta), (pd.tslib.NaTType, type(None)))
def convert_null_or_nat_to_nat(n, **kwargs):
    return pd.NaT


@convert.register(pd.Timedelta, timedelta)
def convert_timedelta_to_pd_timedelta(dt, **kwargs):
    if dt is None:
        return pd.NaT
    return pd.Timedelta(dt)



from __future__ import absolute_import, division, print_function

import pytest
sas7bdat = pytest.importorskip('sas7bdat')
pytest.importorskip('odo.backends.sas')
import os
import pandas as pd
from collections import Iterator
from sas7bdat import SAS7BDAT

from odo.backends.sas import discover, sas_to_iterator
from odo.utils import tmpfile, into_path
from odo import append, convert, resource, dshape


test_path = into_path('backends', 'tests', 'airline.sas7bdat')
sasfile = SAS7BDAT(test_path)


columns = ("DATE", "AIR", "mon1", "mon2", "mon3", "mon4", "mon5", "mon6",
           "mon7", "mon8", "mon9", "mon10", "mon11", "mon12", "t", "Lair")

ds = dshape('''var * {DATE: date, AIR: float64, mon1: float64, mon2: float64,
                      mon3: float64, mon4: float64, mon5: float64,
                      mon6: float64, mon7: float64, mon8: float64,
                      mon9: float64, mon10: float64, mon11: float64,
                      mon12: float64, t: float64, Lair: float64}''')


def test_resource_sas7bdat():
    assert isinstance(resource(test_path), SAS7BDAT)


def test_discover_sas():
    assert discover(sasfile) == ds


def test_convert_sas_to_dataframe():
    df = convert(pd.DataFrame, sasfile)
    assert isinstance(df, pd.DataFrame)

    # pandas doesn't support date
    expected = str(ds.measure).replace('date', 'datetime')

    assert str(discover(df).measure).replace('?', '') == expected


def test_convert_sas_to_list():
    out = convert(list, sasfile)
    assert isinstance(out, list)
    assert not any(isinstance(item, str) for item in out[0])  # No header
    assert all(isinstance(ln, list) for ln in out)


def test_convert_sas_to_iterator():
    itr = sas_to_iterator(sasfile)
    assert isinstance(itr, Iterator)


def test_append_sas_to_sqlite_round_trip():
    expected = convert(set, sasfile)

    with tmpfile('db') as fn:
        r = resource('sqlite:///%s::SAS' % fn, dshape=discover(sasfile))
        append(r, sasfile)

        result = convert(set, r)

    assert expected == result



from __future__ import absolute_import, division, print_function

import pytest
from odo.convert import (convert, list_to_numpy, iterator_to_numpy_chunks,
                         dataframe_to_chunks_dataframe, numpy_to_chunks_numpy,
                         chunks_dataframe_to_dataframe,
                         iterator_to_DataFrame_chunks)
from odo.chunks import chunks
from datashape import discover, dshape
from collections import Iterator
import datetime
import datashape
import numpy as np
import pandas as pd
import pandas.util.testing as tm


def test_basic():
    assert convert(tuple, [1, 2, 3]) == (1, 2, 3)


def test_array_to_set():
    assert convert(set, np.array([1, 2, 3])) == set([1, 2, 3])


def eq(a, b):
    c = a == b
    if isinstance(c, (np.ndarray, pd.Series)):
        c = c.all()
    return c


def test_Series_to_ndarray():
    assert eq(convert(np.ndarray, pd.Series([1, 2, 3]), dshape='3 * float64'),
              np.array([1.0, 2.0, 3.0]))
    assert eq(convert(np.ndarray, pd.Series(['aa', 'bbb', 'ccccc']),
                      dshape='3 * string[5, "A"]'),
              np.array(['aa', 'bbb', 'ccccc'], dtype='S5'))

    assert eq(convert(np.ndarray, pd.Series(['aa', 'bbb', 'ccccc']),
                      dshape='3 * ?string'),
              np.array(['aa', 'bbb', 'ccccc'], dtype='O'))


def test_Series_to_object_ndarray():
    ds = datashape.dshape('{amount: float64, name: string, id: int64}')
    expected = np.array([1.0, 'Alice', 3], dtype='object')
    result = convert(np.ndarray, pd.Series(expected), dshape=ds)
    np.testing.assert_array_equal(result, expected)


def test_Series_to_datetime64_ndarray():
    s = pd.Series(pd.date_range(start='now', freq='N', periods=10).values)
    expected = s.values
    result = convert(np.ndarray, s.values)
    np.testing.assert_array_equal(result, expected)


def test_set_to_Series():
    assert eq(convert(pd.Series, set([1, 2, 3])),
              pd.Series([1, 2, 3]))


def test_Series_to_set():
    assert convert(set, pd.Series([1, 2, 3])) == set([1, 2, 3])


def test_dataframe_and_series():
    s = pd.Series([1, 2, 3], name='foo')
    df = convert(pd.DataFrame, s)
    assert isinstance(df, pd.DataFrame)
    assert list(df.columns) == ['foo']

    s2 = convert(pd.Series, df)
    assert isinstance(s2, pd.Series)

    assert s2.name == 'foo'


def test_iterator_and_numpy_chunks():
    c = iterator_to_numpy_chunks([1, 2, 3], chunksize=2)
    assert isinstance(c, chunks(np.ndarray))
    assert all(isinstance(chunk, np.ndarray) for chunk in c)

    c = iterator_to_numpy_chunks([1, 2, 3], chunksize=2)
    L = convert(list, c)
    assert L == [1, 2, 3]


def test_list_to_numpy():
    ds = datashape.dshape('3 * int32')
    x = list_to_numpy([1, 2, 3], dshape=ds)
    assert (x == [1, 2, 3]).all()
    assert isinstance(x, np.ndarray)

    ds = datashape.dshape('3 * ?int32')
    x = list_to_numpy([1, None, 3], dshape=ds)
    assert np.isnan(x[1])


def test_list_of_single_element_tuples_to_series():
    data = [(1,), (2,), (3,)]
    ds = datashape.dshape('3 * {id: int64}')
    result = convert(pd.Series, data, dshape=ds)
    expected = pd.Series([1, 2, 3], name='id')
    tm.assert_series_equal(result, expected)


def test_cannot_convert_to_series_from_more_than_one_column():
    data = [(1, 2), (2, 3), (3, 4)]
    ds = datashape.dshape('3 * {id: int64, id2: int64}')
    with pytest.raises(ValueError):
        convert(pd.Series, data, dshape=ds)


def test_list_to_numpy_on_tuples():
    data = [['a', 1], ['b', 2], ['c', 3]]
    ds = datashape.dshape('var * (string[1], int32)')
    x = list_to_numpy(data, dshape=ds)
    assert convert(list, x) == [('a', 1), ('b', 2), ('c', 3)]


def test_list_to_numpy_on_dicts():
    data = [{'name': 'Alice', 'amount': 100},
            {'name': 'Bob', 'amount': 200}]
    ds = datashape.dshape('var * {name: string[5], amount: int}')
    x = list_to_numpy(data, dshape=ds)
    assert convert(list, x) == [('Alice', 100), ('Bob', 200)]


def test_list_of_dicts_with_missing_to_numpy():
    data = [{'name': 'Alice', 'amount': 100},
            {'name': 'Bob'},
            {'amount': 200}]
    result = convert(np.ndarray, data)
    assert result.dtype.names == ('amount', 'name')
    expected = np.array([(100.0, 'Alice'),
                         (np.nan, 'Bob'),
                         (200.0, None)],
                        dtype=[('amount', 'float64'), ('name', 'O')])
    assert np.all((result == expected) |
                  ((result != result) & (expected != expected)))


def test_chunks_numpy_pandas():
    x = np.array([('Alice', 100), ('Bob', 200)],
                 dtype=[('name', 'S7'), ('amount', 'i4')])
    n = chunks(np.ndarray)([x, x])

    pan = convert(chunks(pd.DataFrame), n)
    num = convert(chunks(np.ndarray), pan)

    assert isinstance(pan, chunks(pd.DataFrame))
    assert all(isinstance(chunk, pd.DataFrame) for chunk in pan)

    assert isinstance(num, chunks(np.ndarray))
    assert all(isinstance(chunk, np.ndarray) for chunk in num)


def test_numpy_launders_python_types():
    ds = datashape.dshape('3 * int32')
    x = convert(np.ndarray, ['1', '2', '3'], dshape=ds)
    assert convert(list, x) == [1, 2, 3]


def test_numpy_asserts_type_after_dataframe():
    df = pd.DataFrame({'name': ['Alice'], 'amount': [100]})
    ds = datashape.dshape('1 * {name: string[10, "ascii"], amount: int32}')
    x = convert(np.ndarray, df, dshape=ds)
    assert discover(x) == ds


def test_list_to_dataframe_without_datashape():
    data = [('Alice', 100), ('Bob', 200)]
    df = convert(pd.DataFrame, data)
    assert isinstance(df, pd.DataFrame)
    assert list(df.columns) != ['Alice', 100]
    assert convert(list, df) == data


def test_noop():
    assert convert(list, [1, 2, 3]) == [1, 2, 3]


def test_generator_is_iterator():
    g = (1 for i in range(3))
    L = convert(list, g)
    assert L == [1, 1, 1]


def test_list_of_lists_to_set_creates_tuples():
    assert convert(set, [[1], [2]]) == set([(1,), (2,)])


def test_list_of_strings_to_set():
    assert convert(set, ['Alice', 'Bob']) == set(['Alice', 'Bob'])


def test_datetimes_persist():
    typs = [list, tuple, np.ndarray, tuple]
    L = [datetime.datetime.now()] * 3
    ds = discover(L)

    x = L
    for cls in typs:
        x = convert(cls, x)
        assert discover(x) == ds


def test_numpy_to_list_preserves_ns_datetimes():
    x = np.array([(0, 0)], dtype=[('a', 'M8[ns]'), ('b', 'i4')])

    assert convert(list, x) == [(datetime.datetime(1970, 1, 1, 0, 0), 0)]


def test_numpy_to_chunks_numpy():
    x = np.arange(100)
    c = numpy_to_chunks_numpy(x, chunksize=10)
    assert isinstance(c, chunks(np.ndarray))
    assert len(list(c)) == 10
    assert eq(list(c)[0], x[:10])


def test_pandas_and_chunks_pandas():
    df = pd.DataFrame({'a': [1, 2, 3, 4], 'b': [1., 2., 3., 4.]})

    c = dataframe_to_chunks_dataframe(df, chunksize=2)
    assert isinstance(c, chunks(pd.DataFrame))
    assert len(list(c)) == 2

    df2 = chunks_dataframe_to_dataframe(c)
    tm.assert_frame_equal(df, df2)


def test_iterator_to_DataFrame_chunks():
    data = ((0, 1), (2, 3), (4, 5), (6, 7))
    df1 = pd.DataFrame(list(data))
    df2 = iterator_to_DataFrame_chunks(data, chunksize=2, add_index=True)
    df2 = pd.concat(df2, axis=0)
    tm.assert_frame_equal(df1, df2)
    df2 = convert(pd.DataFrame, data, chunksize=2, add_index=True)
    tm.assert_almost_equal(df1, df2)


def test_recarray():
    data = np.array([(1, 1.), (2, 2.)], dtype=[('a', 'i4'), ('b', 'f4')])
    result = convert(np.recarray, data)
    assert isinstance(result, np.recarray)
    assert eq(result.a, data['a'])

    result2 = convert(np.ndarray, data)
    assert not isinstance(result2, np.recarray)
    assert eq(result2, data)


def test_empty_iterator_to_chunks_dataframe():
    ds = dshape('var * {x: int}')
    result = convert(chunks(pd.DataFrame), iter([]), dshape=ds)
    data = convert(pd.DataFrame, result)
    assert isinstance(data, pd.DataFrame)
    assert list(data.columns) == ['x']


def test_empty_iterator_to_chunks_ndarray():
    ds = dshape('var * {x: int}')
    result = convert(chunks(np.ndarray), iter([]), dshape=ds)
    data = convert(np.ndarray, result)
    assert isinstance(data, np.ndarray)
    assert len(data) == 0
    assert data.dtype.names == ('x',)


def test_chunks_of_lists_and_iterators():
    L = [1, 2], [3, 4]
    cl = chunks(list)(L)
    assert convert(list, cl) == [1, 2, 3, 4]
    assert list(convert(Iterator, cl)) == [1, 2, 3, 4]
    assert len(list(convert(chunks(Iterator), cl))) == 2


def test_ndarray_to_df_preserves_field_names():
    ds = dshape('2 * {a: int, b: int}')
    arr = np.array([[0, 1], [2, 3]])
    # dshape explicitly sets field names.
    assert (convert(pd.DataFrame, arr, dshape=ds).columns == ['a', 'b']).all()
    # no dshape is passed.
    assert (convert(pd.DataFrame, arr).columns == [0, 1]).all()


def test_iterator_to_df():
    ds = dshape('var * int32')
    it = iter([1, 2, 3])
    df = convert(pd.DataFrame, it, dshape=ds)
    assert df[0].tolist() == [1, 2, 3]

    it = iter([1, 2, 3])
    df = convert(pd.DataFrame, it, dshape=None)
    assert df[0].tolist() == [1, 2, 3]

    it = iter([1, 2, 3])
    df = convert(pd.DataFrame, it)
    assert df[0].tolist() == [1, 2, 3]



from __future__ import absolute_import, division, print_function
from collections import Iterator

import numpy as np
import pandas as pd
from datashape.dispatch import dispatch
from datashape import from_numpy, var

from dask.array.core import Array, from_array
from dask.bag.core import Bag
import dask.bag as db
from dask.compatibility import long
import dask.dataframe as dd

from odo import append, chunks, convert, discover, TextFile
from ..utils import filter_kwargs


@discover.register(Array)
def discover_dask_array(a, **kwargs):
    return from_numpy(a.shape, a.dtype)


@discover.register(dd.Series)
@discover.register(dd.DataFrame)
def discover_dask_dataframe(df):
    return var * discover(df.head()).measure


arrays = [np.ndarray]

try:
    import h5py
except ImportError:
    pass
else:
    arrays.append(h5py.Dataset)

    @dispatch(h5py.Dataset, (int, long))
    def resize(x, size):
        s = list(x.shape)
        s[0] = size
        return resize(x, tuple(s))

    @dispatch(h5py.Dataset, tuple)
    def resize(x, shape):
        return x.resize(shape)

try:
    import bcolz
except ImportError:
    pass
else:
    arrays.append(bcolz.carray)

    @dispatch(bcolz.carray, (int, long))
    def resize(x, size):
        return x.resize(size)


@convert.register(Array, tuple(arrays), cost=1.)
def array_to_dask(x, name=None, chunks=None, **kwargs):
    if chunks is None:
        raise ValueError("chunks cannot be None")
    return from_array(x, chunks=chunks, name=name,
                      **filter_kwargs(from_array, kwargs))


@convert.register(np.ndarray, Array, cost=10.)
def dask_to_numpy(x, **kwargs):
    return np.array(x)


@convert.register(pd.DataFrame, dd.DataFrame, cost=200)
@convert.register(pd.Series, dd.Series, cost=200)
@convert.register(float, Array, cost=200)
def dask_to_other(x, **kwargs):
    return x.compute()


@append.register(tuple(arrays), Array)
def store_Array_in_ooc_data(out, arr, inplace=False, **kwargs):
    if not inplace:
        # Resize output dataset to accept new data
        assert out.shape[1:] == arr.shape[1:]
        resize(out, out.shape[0] + arr.shape[0])  # elongate
    arr.store(out)
    return out


@convert.register(Iterator, Bag)
def bag_to_iterator(x, **kwargs):
    return iter(x)


@convert.register(Bag, chunks(TextFile))
def bag_to_iterator(x, **kwargs):
    return db.read_text([tf.path for tf in x])


@convert.register(Bag, list)
def bag_to_iterator(x, **kwargs):
    return db.from_sequence(x, **filter_kwargs(db.from_sequence, kwargs))


@convert.register(dd.DataFrame, pd.DataFrame, cost=1.)
def pandas_dataframe_to_dask_dataframe(x, npartitions=None, **kwargs):
    if npartitions is None:
        raise ValueError("npartitions cannot be None")
    return dd.from_pandas(x, npartitions=npartitions,
                          **filter_kwargs(dd.from_pandas, kwargs))



from __future__ import absolute_import, division, print_function

import os
import sys
import subprocess
import uuid
import mmap
from tempfile import NamedTemporaryFile

from contextlib import closing
from functools import partial
from distutils.spawn import find_executable

import pandas as pd
from pandas.formats.format import CSVFormatter
import sqlalchemy as sa
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.elements import Executable, ClauseElement

from toolz import merge
from multipledispatch import MDNotImplementedError

from ..append import append
from ..convert import convert
from ..compatibility import StringIO
from ..utils import literal_compile
from .csv import CSV, infer_header
from ..temp import Temp
from .aws import S3
from .sql import getbind


class CopyFromCSV(Executable, ClauseElement):
    def __init__(self, element, csv, delimiter=',', header=None, na_value='',
                 lineterminator='\n', quotechar='"', escapechar='\\',
                 encoding='utf8', skiprows=0, bind=None, **kwargs):
        if not isinstance(element, sa.Table):
            raise TypeError('element must be a sqlalchemy.Table instance')
        self.element = element
        self.csv = csv
        self.delimiter = delimiter
        self.header = (
            header if header is not None else
            (csv.has_header
             if csv.has_header is not None else infer_header(csv.path))
        )
        self.na_value = na_value
        self.lineterminator = lineterminator
        self.quotechar = quotechar
        self.escapechar = escapechar
        self.encoding = encoding
        self.skiprows = int(skiprows or self.header)
        self._bind = getbind(element, bind)

        for k, v in kwargs.items():
            setattr(self, k, v)

        @property
        def bind(self):
            return self._bind


@compiles(CopyFromCSV, 'sqlite')
def compile_from_csv_sqlite(element, compiler, **kwargs):
    if not find_executable('sqlite3'):
        raise MDNotImplementedError("Could not find sqlite executable")

    t = element.element
    if not element.header:
        csv = element.csv
    else:
        csv = Temp(CSV)('.%s' % uuid.uuid1())
        assert csv.has_header, \
            'SQLAlchemy element.header is True but CSV inferred no header'

        # write to a temporary file after skipping the first line
        chunksize = 1 << 24  # 16 MiB
        lineterminator = element.lineterminator.encode(element.encoding)

        with element.csv.open() as f:
            with closing(mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)) as mf:
                index = mf.find(lineterminator)
                if index == -1:
                    raise ValueError("'%s' not found" % lineterminator)
                mf.seek(index + len(lineterminator))  # len because \r\n
                with open(csv.path, 'wb') as g:
                    for chunk in iter(partial(mf.read, chunksize), b''):
                        g.write(chunk)

    fullpath = os.path.abspath(csv.path).encode('unicode-escape').decode()
    cmd = ['sqlite3',
           '-nullvalue', repr(element.na_value),
           '-separator', element.delimiter,
           '-cmd', '.import "%s" \'%s\'' % (
               # FIXME: format_table(t) is correct, but sqlite will complain
               fullpath, compiler.preparer.format_table(t)
           ),
           element.bind.url.database]
    stderr = subprocess.check_output(
        cmd,
        stderr=subprocess.STDOUT,
        stdin=subprocess.PIPE,
    ).decode(sys.getfilesystemencoding())
    if stderr:
        raise sa.exc.DatabaseError(' '.join(cmd), [], OSError(stderr))
    return ''


@compiles(CopyFromCSV, 'mysql')
def compile_from_csv_mysql(element, compiler, **kwargs):
    if element.na_value:
        raise ValueError(
            'MySQL does not support custom NULL values for CSV input'
        )
    return compiler.process(
        sa.text(
            """LOAD DATA {local} INFILE :path
            INTO TABLE {table}
            CHARACTER SET :encoding
            FIELDS
                TERMINATED BY :delimiter
                ENCLOSED BY :quotechar
                ESCAPED BY :escapechar
            LINES TERMINATED BY :lineterminator
            IGNORE :skiprows LINES
            """.format(
                local=getattr(element, 'local', ''),
                table=compiler.preparer.format_table(element.element)
            )
        ).bindparams(
            path=os.path.abspath(element.csv.path),
            encoding=element.encoding or element.bind.execute(
                'select @@character_set_client'
            ).scalar(),
            delimiter=element.delimiter,
            quotechar=element.quotechar,
            escapechar=element.escapechar,
            lineterminator=element.lineterminator,
            skiprows=int(element.header)
        ),
        **kwargs
    )


@compiles(CopyFromCSV, 'postgresql')
def compile_from_csv_postgres(element, compiler, **kwargs):
    if len(element.escapechar) != 1:
        raise ValueError(
            'postgres does not allow escape characters longer than 1 byte when '
            'bulk loading a CSV file'
        )
    if element.lineterminator != '\n':
        raise ValueError(
            r'PostgreSQL does not support line terminators other than \n'
        )
    return compiler.process(
        sa.text(
            """
            COPY {0} FROM STDIN (
                FORMAT CSV,
                DELIMITER :delimiter,
                NULL :na_value,
                QUOTE :quotechar,
                ESCAPE :escapechar,
                HEADER :header,
                ENCODING :encoding
            )
            """.format(compiler.preparer.format_table(element.element))
        ).bindparams(
            delimiter=element.delimiter,
            na_value=element.na_value,
            quotechar=element.quotechar,
            escapechar=element.escapechar,
            header=element.header,
            encoding=element.encoding or element.bind(
                'show client_encoding'
            ).execute().scalar()
        ),
        **kwargs
    )


try:
    import boto
    from odo.backends.aws import S3
    from redshift_sqlalchemy.dialect import CopyCommand
except ImportError:
    pass
else:
    @compiles(CopyFromCSV, 'redshift')
    def compile_from_csv_redshift(element, compiler, **kwargs):
        assert isinstance(element.csv, S3(CSV))
        assert element.csv.path.startswith('s3://')

        cfg = boto.Config()

        aws_access_key_id = cfg.get('Credentials', 'aws_access_key_id')
        aws_secret_access_key = cfg.get('Credentials', 'aws_secret_access_key')

        compression = getattr(element, 'compression', '').upper() or None
        cmd = CopyCommand(
            table=element.element,
            data_location=element.csv.path,
            access_key_id=aws_access_key_id,
            secret_access_key=aws_secret_access_key,
            format='CSV',
            delimiter=element.delimiter,
            ignore_header=int(element.header),
            empty_as_null=True,
            blanks_as_null=False,
            compression=compression
        )
        return compiler.process(cmd)


@append.register(sa.Table, CSV)
def append_csv_to_sql_table(tbl, csv, bind=None, **kwargs):
    bind = getbind(tbl, bind)
    dialect = bind.dialect.name

    # move things to a temporary S3 bucket if we're using redshift and we
    # aren't already in S3
    if dialect == 'redshift' and not isinstance(csv, S3(CSV)):
        csv = convert(Temp(S3(CSV)), csv, **kwargs)
    elif dialect != 'redshift' and isinstance(csv, S3(CSV)):
        csv = convert(Temp(CSV), csv, has_header=csv.has_header, **kwargs)
    elif dialect == 'hive':
        from .ssh import SSH
        return append(tbl, convert(Temp(SSH(CSV)), csv, **kwargs), **kwargs)

    kwargs = merge(csv.dialect, kwargs)
    stmt = CopyFromCSV(tbl, csv, bind=bind, **kwargs)

    if dialect == 'postgresql':
        with bind.begin() as c:
            with csv.open() as f:
                c.connection.cursor().copy_expert(literal_compile(stmt), f)
    else:
        with bind.begin() as conn:
            conn.execute(stmt)
    return tbl


class NanIsNotNullFormatter(CSVFormatter):
    def _save_chunk(self, start_i, end_i):

        data_index = self.data_index

        # create the data for a chunk
        slicer = slice(start_i, end_i)
        for i in range(len(self.blocks)):
            b = self.blocks[i]
            na_rep = 'nan' if b.dtype.kind == 'f' else self.na_rep
            d = b.to_native_types(slicer=slicer,
                                  na_rep=na_rep,
                                  float_format=self.float_format,
                                  decimal=self.decimal,
                                  date_format=self.date_format,
                                  quoting=self.quoting)

            for col_loc, col in zip(b.mgr_locs, d):
                # self.data is a preallocated list
                self.data[col_loc] = col

        ix = data_index.to_native_types(slicer=slicer, na_rep=self.na_rep,
                                        float_format=self.float_format,
                                        decimal=self.decimal,
                                        date_format=self.date_format,
                                        quoting=self.quoting)

        pd.lib.write_csv_rows(self.data,
                              ix,
                              self.nlevels,
                              self.cols,
                              self.writer)


# The set of sql dialects which do not bounce dataframes to disk to append
# to a sql table.
DATAFRAME_TO_TABLE_IN_MEMORY_CSV = frozenset({
    'postgresql',
})

# TODO: figure out how to dynamically detect this number, users may have
# compiled their sqlite with a different max variable number but this is the
# default.
SQLITE_MAX_VARIABLE_NUMBER = 999


@append.register(sa.Table, pd.DataFrame)
def append_dataframe_to_sql_table(tbl,
                                  df,
                                  bind=None,
                                  dshape=None,
                                  sqlite_max_variable_number=SQLITE_MAX_VARIABLE_NUMBER,
                                  **kwargs):
    bind = getbind(tbl, bind)
    dialect = bind.dialect.name

    if dialect == 'sqlite':
        name = ('.'.join((tbl.schema, tbl.name))
                if tbl.schema is not None else
                tbl.name)
        df.to_sql(
            name,
            bind,
            index=False,
            if_exists='append',
            chunksize=sqlite_max_variable_number,
        )
        return tbl

    if dialect in DATAFRAME_TO_TABLE_IN_MEMORY_CSV:
        buf = StringIO()
        path = None
    else:
        buf = NamedTemporaryFile(mode='w+')
        path = buf.name

    with buf:
        NanIsNotNullFormatter(
            df[[c.name for c in tbl.columns]],
            buf,
            index=False,
        ).save()
        buf.flush()
        buf.seek(0)

        return append_csv_to_sql_table(
            tbl,
            CSV(path, buffer=buf if path is None else None, has_header=True),
            dshape=dshape,
            bind=bind,
            **kwargs
        )



import os
import sys
import pytest

pytestmark = pytest.mark.skipif(sys.platform == 'win32',
                                reason='Requires Mac or Linux')

sa = pytest.importorskip('sqlalchemy')
boto = pytest.importorskip('boto')
pytest.importorskip('psycopg2')
pytest.importorskip('redshift_sqlalchemy')

from contextlib import closing
import json
import itertools
import pandas as pd

import datashape
from odo import into, resource, S3, discover, CSV, drop, odo
from odo.utils import tmpfile
from odo.compatibility import urlopen

with closing(urlopen('http://httpbin.org/ip')) as url:
    public_ip = json.loads(url.read().decode())['origin']
cidrip = public_ip + '/32'

_tmps = ('tmp%d' % i for i in itertools.count())

df = pd.DataFrame({
    'a': list('abc'),
    'b': [1, 2, 3],
    'c': [1.0, 2.0, 3.0]
})[['a', 'b', 'c']]

is_authorized = tried = False


@pytest.fixture(scope='module')
def rs_auth():
    # if we aren't authorized and we've tried to authorize then skip, prevents
    # us from having to deal with timeouts

    # TODO: this will fail if we want to use a testing cluster with a different
    # security group than 'default'
    global is_authorized, tried

    if not is_authorized and not tried:
        if not tried:
            try:
                conn = boto.connect_redshift()
            except boto.exception.NoAuthHandlerFound as e:
                pytest.skip('authorization to access redshift cluster failed '
                            '%s' % e)
            try:
                conn.authorize_cluster_security_group_ingress('default',
                                                              cidrip=cidrip)
            except boto.redshift.exceptions.AuthorizationAlreadyExists:
                is_authorized = True
            except Exception as e:
                pytest.skip('authorization to access redshift cluster failed '
                            '%s' % e)
            else:
                is_authorized = True
            finally:
                tried = True
        else:
            pytest.skip('authorization to access redshift cluster failed')


@pytest.fixture
def db(rs_auth):
    key = os.environ.get('REDSHIFT_DB_URI', None)
    if not key:
        pytest.skip('Please define a non-empty environment variable called '
                    'REDSHIFT_DB_URI to test redshift <- S3')
    else:
        return key


@pytest.yield_fixture
def temp_tb(db):
    t = '%s::%s' % (db, next(_tmps))
    try:
        yield t
    finally:
        drop(resource(t))


@pytest.yield_fixture
def tmpcsv():
    with tmpfile('.csv') as fn:
        with open(fn, mode='w') as f:
            df.to_csv(f, index=False)
        yield fn


tips_uri = 's3://nyqpug/tips.csv'


def test_s3_to_redshift(temp_tb):
    s3 = resource(tips_uri)
    table = into(temp_tb, s3)

    assert discover(table) == discover(s3)
    assert into(set, table) == into(set, s3)


def test_redshift_getting_started(temp_tb):
    dshape = datashape.dshape("""var * {
        userid: int64,
        username: ?string[8],
        firstname: ?string[30],
        lastname: ?string[30],
        city: ?string[30],
        state: ?string[2],
        email: ?string[100],
        phone: ?string[14],
        likesports: ?bool,
        liketheatre: ?bool,
        likeconcerts: ?bool,
        likejazz: ?bool,
        likeclassical: ?bool,
        likeopera: ?bool,
        likerock: ?bool,
        likevegas: ?bool,
        likebroadway: ?bool,
        likemusicals: ?bool,
    }""")
    csv = S3(CSV)('s3://awssampledb/tickit/allusers_pipe.txt')
    table = into(temp_tb, csv, dshape=dshape)

    # make sure we have a non empty table
    assert table.count().scalar() == 49990


def test_redshift_dwdate(temp_tb):
    dshape = datashape.dshape("""var * {
          key: int64,
          date: string[19],
          day_of_week: string[10],
          month: string[10],
          year: int64,
          year_month_num: int64,
          year_month: string[8],
          day_num_in_week: int64,
          day_num_in_month: int64,
          day_num_in_year: int64,
          month_num_in_year: int64,
          week_num_in_year: int64,
          selling_season: string[13],
          last_day_in_week_fl: string[1],
          last_day_in_month_fl: string[1],
          holiday_fl: string[1],
          weekday_fl: string[1]
    }""")
    # we have to pass the separator here because the date column has a comma
    # TODO: see if we can provide a better error message by querying
    # stl_load_errors
    assert odo(S3(CSV)('s3://awssampledb/ssbgz/dwdate'),
               temp_tb,
               delimiter='|',
               compression='gzip',
               dshape=dshape).count().scalar() == 2556


def test_csv_to_redshift(tmpcsv, temp_tb):
    assert into(set, into(temp_tb, tmpcsv)) == into(set, tmpcsv)


def test_frame_to_redshift(temp_tb):
    tb = into(temp_tb, df)
    assert into(set, tb) == into(set, df)



from __future__ import print_function, division, absolute_import

import os
import uuid
import zlib
import re
from fnmatch import fnmatch
from contextlib import contextmanager
from collections import Iterator
from operator import attrgetter
from io import BytesIO

import pandas as pd
from toolz import memoize, first, concat, curry

from .. import (discover, CSV, resource, append, convert, drop, Temp, JSON,
                JSONLines, chunks)

from multipledispatch import MDNotImplementedError

from .text import TextFile

from ..compatibility import urlparse
from ..utils import tmpfile, ext, sample, filter_kwargs, copydoc


@memoize
def get_s3_connection(aws_access_key_id=None,
                      aws_secret_access_key=None,
                      anon=False):
    import boto
    cfg = boto.Config()

    if aws_access_key_id is None:
        aws_access_key_id = cfg.get('Credentials', 'aws_access_key_id')

    if aws_access_key_id is None:
        aws_access_key_id = os.environ.get('AWS_ACCESS_KEY_ID')

    if aws_secret_access_key is None:
        aws_secret_access_key = cfg.get('Credentials', 'aws_secret_access_key')

    if aws_secret_access_key is None:
        aws_secret_access_key = os.environ.get('AWS_SECRET_ACCESS_KEY')

    # anon is False but we didn't provide any credentials so try anonymously
    anon = (not anon and
            aws_access_key_id is None and
            aws_secret_access_key is None)
    return boto.connect_s3(aws_access_key_id, aws_secret_access_key,
                           anon=anon)


class _S3(object):
    """Parametrized S3 bucket Class

    Examples
    --------
    >>> S3(CSV)
    <class 'odo.backends.aws.S3(CSV)'>
    """
    def __init__(self, uri, s3=None, aws_access_key_id=None,
                 aws_secret_access_key=None, *args, **kwargs):
        import boto
        result = urlparse(uri)
        self.bucket = result.netloc
        self.key = result.path.lstrip('/')

        if s3 is not None:
            self.s3 = s3
        else:
            self.s3 = get_s3_connection(aws_access_key_id=aws_access_key_id,
                                        aws_secret_access_key=aws_secret_access_key)
        try:
            bucket = self.s3.get_bucket(self.bucket,
                                        **filter_kwargs(self.s3.get_bucket,
                                                        kwargs))
        except boto.exception.S3ResponseError:
            bucket = self.s3.create_bucket(self.bucket,
                                           **filter_kwargs(self.s3.create_bucket,
                                                           kwargs))

        self.object = bucket.get_key(self.key, **filter_kwargs(bucket.get_key,
                                                               kwargs))
        if self.object is None:
            self.object = bucket.new_key(self.key)

        self.subtype.__init__(self, uri, *args,
                              **filter_kwargs(self.subtype.__init__, kwargs))


@memoize
@copydoc(_S3)
def S3(cls):
    return type('S3(%s)' % cls.__name__, (_S3, cls), {'subtype': cls})


@sample.register((S3(CSV), S3(JSONLines)))
@contextmanager
def sample_s3_line_delimited(data, length=8192):
    """Get a size `length` sample from an S3 CSV or S3 line-delimited JSON.

    Parameters
    ----------
    data : S3(CSV)
        A CSV file living in an S3 bucket
    length : int, optional, default ``8192``
        Number of bytes of the file to read
    """
    headers = {'Range': 'bytes=0-%d' % length}
    if data.object.exists():
        key = data.object
    else:  # we are probably trying to read from a set of files
        keys = sorted(data.object.bucket.list(prefix=data.object.key),
                      key=attrgetter('key'))
        if not keys:
            # we didn't find anything with a prefix of data.object.key
            raise ValueError('Object %r does not exist and no keys with a '
                             'prefix of %r exist' %
                             (data.object, data.object.key))
        key = first(keys)
    raw = key.get_contents_as_string(headers=headers)

    if ext(key.key) == 'gz':
        # decompressobj allows decompression of partial streams
        raw = zlib.decompressobj(32 + zlib.MAX_WBITS).decompress(raw)

    # this is generally cheap as we usually have a tiny amount of data
    try:
        index = raw.rindex(b'\r\n')
    except ValueError:
        index = raw.rindex(b'\n')

    raw = raw[:index]

    with tmpfile(ext(re.sub(r'\.gz$', '', data.path))) as fn:
        # we use wb because without an encoding boto returns bytes
        with open(fn, 'wb') as f:
            f.write(raw)
        yield fn


@discover.register((S3(CSV), S3(JSONLines)))
def discover_s3_line_delimited(c, length=8192, **kwargs):
    """Discover CSV and JSONLines files from S3."""
    with sample(c, length=length) as fn:
        return discover(c.subtype(fn, **kwargs), **kwargs)


@resource.register('s3://.*\.csv(\.gz)?', priority=18)
def resource_s3_csv(uri, **kwargs):
    return S3(CSV)(uri, **kwargs)


@resource.register('s3://.*\*.csv(\.gz)?', priority=19)
def resource_s3_csv_glob(uri, **kwargs):
    con = get_s3_connection()
    result = urlparse(uri)
    bucket = con.get_bucket(result.netloc, validate=False)
    key = result.path.lstrip('/')

    all_keys = bucket.list(prefix=key.split('*')[0])
    matched = [k for k in all_keys if fnmatch(k.key, key)]
    uris = ['s3://{}/{}'.format(bucket.name, k.key) for k in matched]
    r = resource(uris[0], **kwargs)
    return chunks(type(r))([resource(u, **kwargs) for u in sorted(uris)])


@convert.register(chunks(pd.DataFrame), chunks(S3(CSV)), cost=11.0)
def convert_gob_of_s3_csvs_to_chunks_of_dataframes(csvs, **kwargs):
   def _():
       return concat(convert(chunks(pd.DataFrame), csv, **kwargs)
                      for csv in csvs)
   return chunks(pd.DataFrame)(_)


@resource.register('s3://.*\.txt(\.gz)?', priority=18)
def resource_s3_text(uri, **kwargs):
    return S3(TextFile)(uri)


@resource.register('s3://.*\.json(\.gz)?', priority=18)
def resource_s3_json_lines(uri, **kwargs):
    return S3(JSONLines)(uri, **kwargs)


@drop.register((S3(CSV), S3(JSON), S3(JSONLines), S3(TextFile)))
def drop_s3(s3_object):
    s3_object.object.delete()


@drop.register((Temp(S3(CSV)), Temp(S3(JSON)), Temp(S3(JSONLines)),
                Temp(S3(TextFile))))
def drop_temp_s3(s3_object):
    s3_object.object.delete()
    s3_object.object.bucket.delete()


@convert.register(Temp(CSV), (Temp(S3(CSV)), S3(CSV)))
@convert.register(Temp(JSON), (Temp(S3(JSON)), S3(JSON)))
@convert.register(Temp(JSONLines), (Temp(S3(JSONLines)), S3(JSONLines)))
@convert.register(Temp(TextFile), (Temp(S3(TextFile)), S3(TextFile)))
def s3_text_to_temp_text(s3_object, **kwargs):
    tmp_filename = '.%s.%s' % (uuid.uuid1(), ext(s3_object.path))
    s3_object.object.get_contents_to_filename(tmp_filename)
    return Temp(s3_object.subtype)(tmp_filename, **kwargs)


@append.register(CSV, S3(CSV))
@append.register(JSON, S3(JSON))
@append.register(JSONLines, S3(JSONLines))
@append.register(TextFile, S3(TextFile))
def s3_text_to_text(data, s3_object, **kwargs):
    s3_object.object.get_contents_to_filename(data.path)
    return data


@append.register((S3(CSV), Temp(S3(CSV))), (S3(CSV), Temp(S3(CSV))))
@append.register((S3(JSON), Temp(S3(JSON))), (S3(JSON), Temp(S3(JSON))))
@append.register((S3(JSONLines), Temp(S3(JSONLines))),
                 (S3(JSONLines), Temp(S3(JSONLines))))
@append.register((S3(TextFile), Temp(S3(TextFile))),
                 (S3(TextFile), Temp(S3(TextFile))))
def temp_s3_to_s3(a, b, **kwargs):
    a.object.bucket.copy_key(b.object.name, a.object.bucket.name,
                             b.object.name)
    return a


@convert.register(Temp(S3(CSV)), (CSV, Temp(CSV)))
@convert.register(Temp(S3(JSON)), (JSON, Temp(JSON)))
@convert.register(Temp(S3(JSONLines)), (JSONLines, Temp(JSONLines)))
@convert.register(Temp(S3(TextFile)), (TextFile, Temp(TextFile)))
def text_to_temp_s3_text(data, **kwargs):
    subtype = getattr(data, 'persistent_type', type(data))
    uri = 's3://%s/%s.%s' % (uuid.uuid1(), uuid.uuid1(), ext(data.path))
    return append(Temp(S3(subtype))(uri, **kwargs), data)


@append.register((S3(CSV), S3(JSON), S3(JSONLines), S3(TextFile)),
                 (pd.DataFrame, chunks(pd.DataFrame), (list, Iterator)))
def anything_to_s3_text(s3_object, o, **kwargs):
    return append(s3_object, convert(Temp(s3_object.subtype), o, **kwargs), **kwargs)


@contextmanager
def start_multipart_upload_operation(s3_object, **kwargs):
    multipart_upload = s3_object.object.bucket.initiate_multipart_upload(s3_object.key, **kwargs)

    try:
        yield multipart_upload
    except Exception:
        for part in multipart_upload:
            s3_object.object.bucket.cancel_multipart_upload(part.key_name, part.id)
        raise
    else:
        multipart_upload.complete_upload()


@append.register(S3(JSONLines), (JSONLines, Temp(JSONLines)))
@append.register(S3(JSON), (JSON, Temp(JSON)))
@append.register(S3(CSV), (CSV, Temp(CSV)))
@append.register(S3(TextFile), (TextFile, Temp(TextFile)))
def append_text_to_s3(s3_object, data, multipart=False, part_size=5 << 20, **kwargs):
    if multipart:
        filtered_kwargs = filter_kwargs(s3_object.object.bucket.initiate_multipart_upload, kwargs)
        with start_multipart_upload_operation(s3_object, **filtered_kwargs) as multipart_upload:
            with open(data.path, 'rb') as f:
                parts = enumerate(iter(curry(f.read, part_size), ''), start=1)
                for part_num, part in parts:
                    multipart_upload.upload_part_from_file(
                        BytesIO(part),
                        part_num=part_num
                    )
        return s3_object

    s3_object.object.set_contents_from_filename(
        data.path, **filter_kwargs(s3_object.object.set_contents_from_filename, kwargs))
    return s3_object


try:
    from .hdfs import HDFS
except ImportError:
    pass
else:
    @append.register(S3(JSON), HDFS(JSON))
    @append.register(S3(JSONLines), HDFS(JSONLines))
    @append.register(S3(CSV), HDFS(CSV))
    @append.register(S3(TextFile), HDFS(TextFile))
    @append.register(HDFS(JSON), S3(JSON))
    @append.register(HDFS(JSONLines), S3(JSONLines))
    @append.register(HDFS(CSV), S3(CSV))
    @append.register(HDFS(TextFile), S3(TextFile))
    def other_remote_text_to_s3_text(a, b, **kwargs):
        raise MDNotImplementedError()


try:
    from .ssh import connect, _SSH, SSH
except ImportError:
    pass
else:
    @append.register(S3(JSON), SSH(JSON))
    @append.register(S3(JSONLines), SSH(JSONLines))
    @append.register(S3(CSV), SSH(CSV))
    @append.register(S3(TextFile), SSH(TextFile))
    def remote_text_to_s3_text(a, b, **kwargs):
        return append(a, convert(Temp(b.subtype), b, **kwargs), **kwargs)


    @append.register(_SSH, _S3)
    def s3_to_ssh(ssh, s3_object, url_timeout=600, **kwargs):
        if s3_object.s3.anon:
            url = 'https://%s.s3.amazonaws.com/%s' % (s3_object.bucket, s3_object.object.name)
        else:
            url = s3_object.object.generate_url(url_timeout)
        command = "wget '%s' -qO- >> '%s'" % (url, ssh.path)
        conn = connect(**ssh.auth)
        _, stdout, stderr = conn.exec_command(command)
        exit_status = stdout.channel.recv_exit_status()
        if exit_status:
            raise ValueError('Error code %d, message: %r' % (exit_status,
                                                             stderr.read()))
        return ssh



from __future__ import absolute_import, division, print_function

import pandas as pd

from ..convert import convert
from ..utils import ignoring


with ignoring(ImportError):
    from bokeh.models import ColumnDataSource


    @convert.register(pd.DataFrame, ColumnDataSource)
    def columndatasource_to_dataframe(cds, **kwargs):
        df = cds.to_df()
        return df[sorted(df.columns)]


    @convert.register(ColumnDataSource, pd.DataFrame)
    def dataframe_to_columndatasource(df, **kwargs):
        d = ColumnDataSource.from_df(df)
        if 'index' in d:
            d.pop('index')
        return ColumnDataSource(d)



from __future__ import absolute_import, division, print_function

import re


_pattern_type = type(re.compile(''))


def normalize(r):
    """Normalize a regular expression by ensuring that it is wrapped with:
    '^' and '$'

    Parameters
    ----------
    r : str or Pattern
        The pattern to normalize.

    Returns
    -------
    p : Pattern
        The compiled regex.
    """
    if isinstance(r, _pattern_type):
        r = r.pattern
    return re.compile('^' + r.lstrip('^').rstrip('$') + '$')


class RegexDispatcher(object):
    """
    Regular Expression Dispatcher

    >>> f = RegexDispatcher('f')

    >>> f.register('\d*')
    ... def parse_int(s):
    ...     return int(s)

    >>> f.register('\d*\.\d*')
    ... def parse_float(s):
    ...     return float(s)

    Set priorities to break ties between multiple matches.
    Default priority is set to 10

    >>> f.register('\w*', priority=9)
    ... def parse_str(s):
    ...     return s

    >>> type(f('123'))
    int

    >>> type(f('123.456'))
    float
    """
    def __init__(self, name):
        self.name = name
        self.funcs = {}
        self.priorities = {}

    def add(self, regex, func, priority=10):
        self.funcs[normalize(regex)] = func
        self.priorities[func] = priority

    def register(self, regex, priority=10):
        """Register a new handler in this regex dispatcher.

        Parameters
        ----------
        regex : str or Pattern
            The pattern to match against.
        priority : int, optional
            The priority for this pattern. This is used to resolve ambigious
            matches. The highest priority match wins.

        Returns
        -------
        decorator : callable
            A decorator that registers the function with this RegexDispatcher
            but otherwise returns the function unchanged.
        """
        def _(func):
            self.add(regex, func, priority)
            return func
        return _

    def dispatch(self, s):
        funcs = (func for r, func in self.funcs.items() if r.match(s))
        return max(funcs, key=self.priorities.get)

    def __call__(self, s, *args, **kwargs):
        return self.dispatch(s)(s, *args, **kwargs)

    @property
    def __doc__(self):
        # take the min to give the docstring of the last fallback function
        return min(self.priorities.items(), key=lambda x: x[1])[0].__doc__



from __future__ import absolute_import, division, print_function

import numpy as np
from datashape import dshape, DataShape, Option, DateTime, string, TimeDelta
from datashape import Date, to_numpy_dtype, Tuple, String, Decimal
from datashape.predicates import isscalar, isnumeric, isrecord


def unit_to_dtype(ds):
    """ Convert a datashape Unit instance into a numpy dtype

    Parameters
    ----------
    ds : DataShape
        The DataShape instance to convert

    Returns
    -------
    np.dtype

    Examples
    --------
    >>> unit_to_dtype('int32')
    dtype('int32')
    >>> unit_to_dtype('float64')
    dtype('float64')
    >>> unit_to_dtype('?int64')
    dtype('float64')
    >>> unit_to_dtype('string')
    dtype('O')
    >>> unit_to_dtype('?datetime')
    dtype('<M8[us]')
    """
    if isinstance(ds, str):
        ds = dshape(ds)
    if isinstance(ds, DataShape):
        ds = ds.measure
    if isinstance(ds, Option) and isscalar(ds) and isnumeric(ds):
        if isinstance(ds.ty, Decimal):
            return unit_to_dtype(
                str(ds.ty.to_numpy_dtype()).replace('int', 'float')
            )
        return unit_to_dtype(str(ds).replace('int', 'float').replace('?', ''))
    if isinstance(ds, Option) and isinstance(
        ds.ty, (Date, DateTime, String, TimeDelta)
    ):
        ds = ds.ty
    if ds == string:
        return np.dtype('O')
    return to_numpy_dtype(ds)


def dshape_to_numpy(ds):
    """ Convert a datashape to a NumPy dtype

    Parameters
    ----------
    ds : DataShape
        The DataShape instance to convert

    Returns
    -------
    np.dtype

    Examples
    --------
    >>> dshape_to_numpy('int32')
    dtype('int32')
    >>> dshape_to_numpy('?int32')
    dtype('float32')

    >>> dshape_to_numpy('{name: string[5, "ascii"], amount: ?int32}')
    dtype([('name', 'S5'), ('amount', '<f4')])

    >>> dshape_to_numpy('(int32, float32)')
    dtype([('f0', '<i4'), ('f1', '<f4')])
    """
    if isinstance(ds, str):
        ds = dshape(ds)
    if isinstance(ds, DataShape):
        ds = ds.measure
    if isrecord(ds):
        return np.dtype([
            (str(name), unit_to_dtype(typ))
            for name, typ in zip(ds.names, ds.types)
        ])
    if isinstance(ds, Tuple):
        return np.dtype([
            ('f%d' % i, unit_to_dtype(typ))
            for i, typ in enumerate(ds.parameters[0])
        ])
    else:
        return unit_to_dtype(ds)


def dshape_to_pandas(ds):
    """ Convert a datashape to a pair of
    ``({name1: dtype1, name2: dtype2, ...}, [datecol1, datecol2, ...])``

    Parameters
    ----------
    ds : DataShape
        The DataShape instance to convert

    Returns
    -------
    ({str: np.dtype}, [str])

    Examples
    --------
    >>> dshape_to_pandas('{a: int32}')  # doctest: +SKIP
    ({'a': dtype('int32')}, [])

    >>> dshape_to_pandas('{a: int32, when: datetime}')  # doctest: +SKIP
    ({'a': dtype('int32')}, ['when'])

    >>> dshape_to_pandas('{a: ?int64}')  # doctest: +SKIP
    ({'a': dtype('float64')}, [])
    """
    if isinstance(ds, str):
        ds = dshape(ds)
    if isinstance(ds, DataShape) and len(ds) == 1:
        ds = ds[0]

    dtypes = {
        name: (
            np.dtype('object')
            if isinstance(typ, String) else unit_to_dtype(typ)
        )
        for name, typ in ds.measure.dict.items() if 'date' not in str(typ)
    }
    datetimes = [
        name for name, typ in ds.measure.dict.items() if 'date' in str(typ)
    ]

    return dtypes, datetimes



from __future__ import division, print_function, absolute_import

import os
import glob
import itertools
import tempfile
import shutil

from functools import partial
from collections import Iterator
from datetime import datetime, date

import pandas as pd

import toolz
from toolz.curried import get, map, memoize
from toolz import pipe, concat, curry

from pyspark import RDD, SQLContext, HiveContext
from pyspark.sql import SchemaRDD
from pyspark.rdd import PipelinedRDD

import datashape
from datashape import dshape, Record, DataShape, Option, Tuple
from datashape.predicates import isdimension, isrecord, iscollection

from .. import append, discover, convert
from ..core import ooc_types
from ..directory import Directory
from ..temp import Temp
from ..chunks import chunks

from .json import JSONLines, JSON
from .csv import CSV


from pyspark.sql import DataFrame as SparkDataFrame
from pyspark.sql.types import (
    ByteType, ShortType, IntegerType, LongType, FloatType, DoubleType,
    StringType, BinaryType, BooleanType, TimestampType, DateType, ArrayType,
    StructType, StructField
)

base = int, float, datetime, date, bool, str
_names = ('tmp%d' % i for i in itertools.count())


@append.register(SQLContext, object)
def iterable_to_sql_context(ctx, seq, **kwargs):
    return append(ctx, append(ctx._sc, seq, **kwargs), **kwargs)


def register_table(ctx, srdd, name=None):
    if name is None:
        name = next(_names)
    ctx.registerDataFrameAsTable(srdd, name)


@append.register(SQLContext, (JSONLines, Directory(JSONLines)))
def jsonlines_to_sparksql(ctx, json, dshape=None, name=None, schema=None,
                          samplingRatio=0.25, **kwargs):
    # if we're passing in schema, assume that we know what we're doing and
    # bypass any automated dshape inference
    if dshape is not None and schema is None:
        schema = dshape_to_schema(dshape.measure
                                  if isrecord(dshape.measure) else dshape)
    srdd = ctx.jsonFile(json.path, schema=schema, samplingRatio=samplingRatio)
    register_table(ctx, srdd, name=name)
    return srdd


@convert.register(list, (SparkDataFrame, SchemaRDD), cost=200.0)
def sparksql_dataframe_to_list(df, dshape=None, **kwargs):
    result = df.collect()
    if (dshape is not None and iscollection(dshape) and
            not isrecord(dshape.measure)):
        return list(map(get(0), result))
    return result


@convert.register(base, (SparkDataFrame, SchemaRDD), cost=200.0)
def spark_df_to_base(df, **kwargs):
    return df.collect()[0][0]


@append.register(SQLContext, RDD)
def rdd_to_sqlcontext(ctx, rdd, name=None, dshape=None, **kwargs):
    """ Convert a normal PySpark RDD to a SparkSQL RDD or Spark DataFrame

    Schema inferred by ds_to_sparksql.  Can also specify it explicitly with
    schema keyword argument.
    """
    # TODO: assumes that we don't have e.g., 10 * 10 * {x: int, y: int}
    if isdimension(dshape.parameters[0]):
        dshape = dshape.measure
    sql_schema = dshape_to_schema(dshape)
    sdf = ctx.applySchema(rdd, sql_schema)
    if name is None:
        name = next(_names)
    register_table(ctx, sdf, name=name)
    ctx.cacheTable(name)
    return sdf


def scala_set_to_set(ctx, x):
    from py4j.java_gateway import java_import

    # import scala
    java_import(ctx._jvm, 'scala')

    # grab Scala's set converter and convert to a Python set
    return set(ctx._jvm.scala.collection.JavaConversions.setAsJavaSet(x))


@discover.register(SQLContext)
def discover_sqlcontext(ctx):
    table_names = sorted(map(str, ctx.tableNames()))
    dshapes = zip(table_names, map(discover, map(ctx.table, table_names)))
    return datashape.DataShape(datashape.Record(dshapes))


@discover.register((SparkDataFrame, SchemaRDD))
def discover_spark_data_frame(df):
    schema = df.schema() if callable(df.schema) else df.schema
    return datashape.var * schema_to_dshape(schema)


def chunk_file(filename, chunksize):
    """Stream `filename` in chunks of size `chunksize`.

    Parameters
    ----------
    filename : str
        File to chunk
    chunksize : int
        Number of bytes to hold in memory at a single time
    """
    with open(filename, mode='rb') as f:
        for chunk in iter(partial(f.read, chunksize), b''):
            yield chunk


@append.register(JSONLines, (SparkDataFrame, SchemaRDD))
def spark_df_to_jsonlines(js, df,
                          pattern='part-*', chunksize=1 << 23,  # 8MB
                          **kwargs):
    tmpd = tempfile.mkdtemp()
    try:
        try:
            df.save(tmpd, source='org.apache.spark.sql.json', mode='overwrite')
        except AttributeError:
            shutil.rmtree(tmpd)
            df.toJSON().saveAsTextFile(tmpd)
    except:
        raise
    else:
        files = glob.glob(os.path.join(tmpd, pattern))
        with open(js.path, mode='ab') as f:
            pipe(files,
                 map(curry(chunk_file, chunksize=chunksize)),
                 concat,
                 map(f.write),
                 toolz.count)
    finally:
        shutil.rmtree(tmpd)
    return js


@convert.register((SparkDataFrame, SchemaRDD), (RDD, PipelinedRDD))
def rdd_to_spark_df_or_srdd(rdd, **kwargs):
    return append(HiveContext(rdd.context), rdd, **kwargs)


try:
    from .hdfs import HDFS
except ImportError:
    pass
else:
    @append.register(HDFS(JSONLines),
                     (Iterator, object, SparkDataFrame, SchemaRDD))
    @append.register(HDFS(JSON), (list, object))
    @append.register(HDFS(CSV), (chunks(pd.DataFrame), pd.DataFrame, object))
    def append_spark_to_hdfs(target, source, **kwargs):
        tmp = convert(Temp(target.subtype), source, **kwargs)
        return append(target, tmp, **kwargs)


def dshape_to_schema(ds):
    """Convert datashape to SparkSQL type system.

    Examples
    --------
    >>> print(dshape_to_schema('int32'))  # doctest: +SKIP
    IntegerType
    >>> print(dshape_to_schema('5 * int32')  # doctest: +SKIP
    ArrayType(IntegerType,false)
    >>> print(dshape_to_schema('5 * ?int32'))  # doctest: +SKIP
    ArrayType(IntegerType,true)
    >>> print(dshape_to_schema('{name: string, amount: int32}'))  # doctest: +SKIP
    StructType(List(StructField(name,StringType,false),StructField(amount,IntegerType,false)  # doctest: +SKIP))
    >>> print(dshape_to_schema('10 * {name: string, amount: ?int32}'))  # doctest: +SKIP
    ArrayType(StructType(List(StructField(name,StringType,false),StructField(amount,IntegerType,true))),false)
    """
    if isinstance(ds, str):
        return dshape_to_schema(dshape(ds))
    if isinstance(ds, Tuple):
        raise TypeError('Please provide a Record dshape for these column '
                        'types: %s' % (ds.dshapes,))
    if isinstance(ds, Record):
        return StructType([
            StructField(name,
                        dshape_to_schema(deoption(typ)),
                        isinstance(typ, datashape.Option))
            for name, typ in ds.fields])
    if isinstance(ds, DataShape):
        if isdimension(ds[0]):
            elem = ds.subshape[0]
            if isinstance(elem, DataShape) and len(elem) == 1:
                elem = elem[0]
            return ArrayType(dshape_to_schema(deoption(elem)),
                             isinstance(elem, Option))
        else:
            return dshape_to_schema(ds[0])
    if ds in dshape_to_sparksql:
        return dshape_to_sparksql[ds]
    raise NotImplementedError()


def schema_to_dshape(schema):
    if type(schema) in sparksql_to_dshape:
        return sparksql_to_dshape[type(schema)]
    if isinstance(schema, ArrayType):
        dshape = schema_to_dshape(schema.elementType)
        return datashape.var * (Option(dshape)
                                if schema.containsNull else dshape)
    if isinstance(schema, StructType):
        fields = [(field.name, Option(schema_to_dshape(field.dataType))
                  if field.nullable else schema_to_dshape(field.dataType))
                  for field in schema.fields]
        return datashape.dshape(Record(fields))
    raise NotImplementedError('SparkSQL type not known %r' %
                              type(schema).__name__)


def deoption(ds):
    """

    >>> deoption('int32')
    ctype("int32")

    >>> deoption('?int32')
    ctype("int32")
    """
    if isinstance(ds, str):
        ds = dshape(ds)
    if isinstance(ds, DataShape) and not isdimension(ds[0]):
        return deoption(ds[0])
    if isinstance(ds, Option):
        return ds.ty
    else:
        return ds


# see http://spark.apache.org/docs/latest/sql-programming-guide.html#spark-sql-datatype-reference
sparksql_to_dshape = {
    ByteType: datashape.int8,
    ShortType: datashape.int16,
    IntegerType: datashape.int32,
    LongType: datashape.int64,
    FloatType: datashape.float32,
    DoubleType: datashape.float64,
    StringType: datashape.string,
    BinaryType: datashape.bytes_,
    BooleanType: datashape.bool_,
    TimestampType: datashape.datetime_,
    DateType: datashape.date_,
    # sql.ArrayType: ?,
    # sql.MapTYpe: ?,
    # sql.StructType: ?
}

dshape_to_sparksql = {
    datashape.int16: ShortType(),
    datashape.int32: IntegerType(),
    datashape.int64: LongType(),
    datashape.float32: FloatType(),
    datashape.float64: DoubleType(),
    datashape.real: DoubleType(),
    datashape.time_: TimestampType(),
    datashape.date_: DateType(),
    datashape.datetime_: TimestampType(),
    datashape.bool_: BooleanType(),
    datashape.string: StringType()
}

ooc_types |= set([SparkDataFrame, SchemaRDD])

SQLContext = memoize(SQLContext)
HiveContext = memoize(HiveContext)
